#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains tpDcc-tools-datalibrary server implementation for 3ds Max
"""

from __future__ import print_function, division, absolute_import

import os

from tpDcc.core import server
from tpDcc.libs.python import path as path_utils

from tpDcc.libs.datalibrary.core import datalib


class DataLibraryServer(server.DccServer, object):

    PORT = 28231

    def __init__(self, *args, **kwargs):
        super(DataLibraryServer, self).__init__(*args, **kwargs)

        self._data_library = None

    def save_data(self, data, reply):
        library_path = data['library_path']
        data_path = data['data_path']
        values = data['values']

        if not library_path or not os.path.isfile(library_path):
            reply['success'] = False
            reply['msg'] = 'Impossible to save data "{}" because library path "{}" does not exist!'.format(
                data_path, library_path)
            return

        data_lib = self._get_data_library(library_path)
        data_item = data_lib.get(data_path, only_extension=True)
        if not data_item:
            reply['success'] = False
            reply['msg'] = 'Impossible to retrieve data "{}" from data library: "{}"!'.format(data_path, data_lib)
            return

        save_function = data_item.functionality().get('save')
        if not save_function:
            reply['success'] = False
            reply['msg'] = 'Save functionality is not available for data: "{}"'.format(data_item)
            return

        result = save_function(**values)

        reply['success'] = True
        reply['result'] = result

    def _get_data_library(self, library_path):
        if self._data_library and path_utils.clean_path(
                self._data_library.identifier) == path_utils.clean_path(library_path):
            return self._data_library

        self._data_library = datalib.DataLibrary.load(library_path)

        return self._data_library

    def load_data(self, data, reply):
        library_path = data['library_path']
        data_path = data['data_path']

        if not library_path or not os.path.isfile(library_path):
            reply['success'] = False
            reply['message'] = 'Impossible to load data "{}" because library path "{}" does not exist!'.format(
                data_path, library_path)
            return

        if not os.path.isfile(data_path):
            reply['success'] = False
            reply['message'] = 'Impossible to load data "{}" because it does not exist!'.format(data_path)
            return

        data_lib = self._get_data_library(library_path)
        data_item = data_lib.get(data_path)
        if not data_item:
            reply['success'] = False
            reply['message'] = 'Impossible to retrieve data "{}" from data library: "{}"!'.format(data_path, data_lib)
            return

        load_function = data_item.functionality().get('load')
        if not load_function:
            reply['success'] = False
            reply['message'] = 'Load functionality is not available for data: "{}"'.format(data_item)
            return

        load_function()

        reply['success'] = True



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains data library filter menu widget implementation
"""

from __future__ import print_function, division, absolute_import

from functools import partial

from Qt.QtCore import Qt
from Qt.QtWidgets import QFrame, QMenu, QWidgetAction
from Qt.QtGui import QCursor

from tpDcc.libs.qt.core import qtutils
from tpDcc.libs.qt.widgets import layouts, label, checkbox, action


class SortByMenu(QMenu):
    def __init__(self, *args, **kwargs):
        super(SortByMenu, self).__init__(*args, **kwargs)

        self._library = None

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def set_library(self, library):
        self._library = library


class GroupByMenu(QMenu):
    def __init__(self, *args, **kwargs):
        super(GroupByMenu, self).__init__(*args, **kwargs)

        self._library = None

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def set_library(self, library):
        self._library = library


class FilterByMenu(QMenu):
    def __init__(self, *args, **kwargs):
        super(FilterByMenu, self).__init__(*args, **kwargs)

        self._library = None
        self._facets = list()
        self._options = {'field': 'type'}
        self._settings = dict()

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def show(self, point=None):
        """
        Overrides base QMeu show function to allow to show it in a specific position
        :param point: QPoint or None
        """

        self.clear()

        field = self._options.get('field')
        queries = self.library().queries(exclude=self.name())
        self._facets = self.library().distinct(field, queries=queries)

        show_action = action.SeparatorAction('Show {}'.format(field.title()), self)
        self.addAction(show_action)

        show_all_action = action.LabelAction('Show All', self)
        self.addAction(show_all_action)
        show_all_action.setEnabled(not self.is_show_all_enabled())
        show_all_callback = partial(self._on_show_all_action_clicked)
        show_all_action.triggered.connect(show_all_callback)

        self.addSeparator()

        for facet in self._facets:
            name = facet.get('name', '')
            checked = self.settings().get(name, True)
            filter_by_action = FilterByAction(self)
            filter_by_action.set_facet(facet)
            filter_by_action.setChecked(checked)
            self.addAction(filter_by_action)
            callback = partial(self._on_action_checked, name, not checked)
            filter_by_action.triggered.connect(callback)

        point = point or QCursor.pos()

        self.exec_(point)

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def library(self):
        """
        Returns the library model for the menu
        :return: Library
        """

        return self._library

    def set_library(self, library):
        """
        Set the library model for the menu
        :param library: Library
        """

        if library == self._library:
            return

        if not library and self._library:
            try:
                qtutils.safe_disconnect_signal(self._library.searchStarted)
            except Exception as exc:
                pass

        self._library = library

        if self._library:
            self._library.searchStarted.connect(self._on_search_init)

    def name(self):
        """
        Returns the name of the fulter used by the library
        """

        return self._options.get('field') + 'FilterMenu'

    def is_active(self):
        """
        Returns whether are any filters currently active using the settings
        :return: bool
        """

        settings = self.settings()
        for name in settings:
            if not settings.get(name):
                return True

        return False

    def is_show_all_enabled(self):
        """
        Returns whether all current filters are enabled or not
        :return: bool
        """

        for facet in self._facets:
            if not self._settings.get(facet.get('name'), True):
                return False

        return True

    def set_all_enabled(self, enabled):
        """
        Set all the filters enabled
        :param enabled: bool
        """

        for facet in self._facets:
            self._settings[facet.get('name')] = enabled

    def set_options(self, options):
        """
        Sets the options to be used by the filters menu
        :param options: dict
        """

        self._options = options

    def settings(self):
        """
        Returns the settings for the filter menu
        :return: dict
        """

        return self._settings

    def set_settings(self, settings):
        """
        Set the settings for the filter menu
        :param settings: dict
        """

        self._settings = settings

    # ============================================================================================================
    # CALLBACKS
    # ============================================================================================================

    def _on_search_init(self):
        """
        Internal callback function that is called before each search to update the filter men query
        """

        library = self.library()
        if not library:
            return

        filters = list()

        settings = self.settings()
        field = self._options.get('field')
        for name in settings:
            checked = settings.get(name, True)
            if not checked:
                filters.append((field, 'not', name))

        query = {
            'name': self.name(),
            'operator': 'and',
            'filters': filters
        }

        self.library().add_query(query)

    def _on_show_all_action_clicked(self):
        """
        Internal callback function that is triggered when the user clicks the show all action
        """

        self.set_all_enabled(True)
        self.library().search()

    def _on_action_checked(self, name, checked):
        """
        Internal callback function triggered when an action has been clicked
        :param name: str
        :param checked: bool
        """

        if qtutils.is_control_modifier():
            self.set_all_enabled(False)
            self._settings[name] = True
        else:
            self._settings[name] = checked

        self.library().search()


class FilterByAction(QWidgetAction, object):
    def __init__(self, parent=None):
        super(FilterByAction, self).__init__(parent)

        self._facet = None
        self._checked = False

    def setChecked(self, checked):
        """
        Overrides base QWidgetAction setChecked function
        :param checked: bool
        """

        self._checked = checked

    def set_facet(self, facet):
        """
        Sets action facet
        :param facet:
        """

        self._facet = facet

    def createWidget(self, menu):
        """
        Overrides base QWidgetAction createWidget function
        :param menu: QMenu
        :return: QWidget
        """

        widget = QFrame(self.parent())
        widget.setObjectName('filterByAction')
        facet = self._facet
        name = facet.get('name', '')
        count = str(facet.get('count', 0))
        title = name.replace('.', '').title()
        cbx = checkbox.BaseCheckBox(parent=widget)
        cbx.setAttribute(Qt.WA_TransparentForMouseEvents)
        cbx.setText(title)
        cbx.installEventFilter(self)
        cbx.setChecked(self._checked)
        label2 = label.BaseLabel(parent=widget)
        label2.setObjectName('actionCounter')
        label2.setText(count)
        layout = layouts.HorizontalLayout(spacing=0, margins=(0, 0, 0, 0))
        layout.addWidget(cbx, stretch=1)
        layout.addWidget(label2)
        widget.setLayout(layout)

        cbx.toggled.connect(self._on_triggered)

        return widget

    def _on_triggered(self, checked=None):
        """
        Triggered when teh checkbox value has changed
        :param checked: bool
        """

        self.triggered.emit()
        self.parent().close()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains library data item implementation for Maya
"""

from __future__ import print_function, division, absolute_import

import logging

from Qt.QtWidgets import QAction
from Qt.QtGui import QCursor

from tpDcc.managers import resources
from tpDcc.tools.datalibrary.data import base

LOGGER = logging.getLogger('tpDcc-tools-datalibrary')


class MayaDataItemView(base.BaseDataItemView):
    def __init__(self, *args, **kwargs):
        super(MayaDataItemView, self).__init__(*args, **kwargs)

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def context_menu(self, menu, items=None):
        """
        Returns the context menu for the item
        :return: QMenu
        """

        super(MayaDataItemView, self).context_menu(menu, items=items)

        clean_student_license_function = self.item.functionality().get('clean_student_license')
        if clean_student_license_function:
            clean_student_license_action = QAction(resources.icon('student'), 'Clean Student License', menu)
            menu.addAction(clean_student_license_action)
            clean_student_license_action.triggered.connect(lambda: clean_student_license_function())

        menu.addSeparator()
        create_selection_sets_menu = self._create_selection_sets_menu(menu, enable_select_content=False)
        menu.insertMenu(self._select_content_action, create_selection_sets_menu)

    def select_content(self, namespaces=None, **kwargs):
        """
        Select the contents of this item in the current DCC scene
        :param namespaces: list(str) or None
        :param kwargs: dict
        """

        namespaces = namespaces or self.namespaces()
        kwargs = kwargs or self._selection_modifiers()
        msg = 'Select content: Item.selectContent(namespaces={}, kwargs={})'.format(namespaces, kwargs)
        LOGGER.debug(msg)

        try:
            self.transfer_object().select(namespaces=namespaces, **kwargs)
        except Exception as exc:
            self.show_error_dialog('Item Error', str(exc))
            raise

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def namespaces(self):
        """
        Returns the namespaces for this item depending on the namespace option
        :return: list(str) or None
        """

        return list()

        # return self.current_load_value('namespaces')

    def namespace_option(self):
        """
        Returns the namespace option for this item
        :return: NamespaceOption or None
        """

        return self.current_load_value('namespaceOption')

    def show_selection_sets_menu(self, **kwargs):
        """
        Shows the selection sets menu for this item at the current cursor position
        :param kwargs: dict
        :return: QAction
        """

        menu = self._create_selection_sets_menu(**kwargs)
        position = QCursor.pos()

        return menu.exec_(position)

    # ============================================================================================================
    # INTERNAL
    # ============================================================================================================

    def _create_selection_sets_menu(self, parent=None, enable_select_content=True):
        """
        Internal function that crates a new instance of the selection sets menu
        :param parent: QWidget
        :param enable_select_content:, bool
        :return: QMenu
        """

        # Import here to avoid cycles in Python 2
        from tpDcc.tools.datalibrary.dccs.maya.menus import sets

        parent = parent or self.library_window()
        namespaces = self.namespaces()
        menu = sets.SetsMenu(
            item=self, parent=parent, namespaces=namespaces, enable_select_content=enable_select_content)

        return menu



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Tool to manage DCC related data in an easy way.
"""

from __future__ import print_function, division, absolute_import

import os
import sys
import logging

from tpDcc.core import tool
from tpDcc.core import dcc

from tpDcc.tools.datalibrary.core import consts, client, toolset

LOGGER = logging.getLogger(consts.TOOL_ID)


class DataLibraryTool(tool.DccTool, object):

    ID = consts.TOOL_ID
    CLIENT_CLASS = client.DataLibraryClient
    TOOLSET_CLASS = toolset.DataLibraryToolset

    def __init__(self, *args, **kwargs):
        super(DataLibraryTool, self).__init__(*args, **kwargs)

    @classmethod
    def config_dict(cls, file_name=None):
        base_tool_config = tool.DccTool.config_dict(file_name=file_name)
        tool_config = {
            'name': 'Data Library',
            'id': cls.ID,
            'supported_dccs': {
                dcc.Dccs.Maya: ['2017', '2018', '2019', '2020'],
                dcc.Dccs.Max: ['2017.0', '2018.0', '2019.0', '2020.0']
            },
            'icon': 'datalibrary',
            'tooltip': 'Tool to manage DCC related data in an easy way.',
            'tags': ['tpDcc', 'dcc', 'tool', 'data'],
            'is_checkable': False,
            'is_checked': False,
            'menu_ui': {'label': 'Data Library', 'load_on_startup': False, 'color': '', 'background_color': ''},
        }
        base_tool_config.update(tool_config)

        return base_tool_config

    def launch(self, *args, **kwargs):
        return self.launch_frameless(*args, **kwargs)


if __name__ == '__main__':
    import tpDcc.loader
    from tpDcc.managers import tools

    tool_path = os.path.abspath(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__)))))
    if tool_path not in sys.path:
        sys.path.append(tool_path)

    tpDcc.loader.init()
    tools.ToolsManager().launch_tool_by_id(DataLibraryTool.ID)
    # tools.ToolsManager().launch_tool_by_id(consts.TOOL_ID)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains base item factory implementation
"""

from __future__ import print_function, division, absolute_import

import os

from tpDcc import dcc
from tpDcc.libs.python import decorators, plugin
from tpDcc.libs.datalibrary.data import folder

from tpDcc.tools.datalibrary.core.views import item
from tpDcc.tools.datalibrary.data import base, folder as folder_view
from tpDcc.tools.datalibrary.widgets import load, save, export


class _MetaItemsFactory(type):

    def __call__(self, *args, **kwargs):
        return type.__call__(BaseItemsFactory, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        if dcc.client().is_maya():
            from tpDcc.tools.datalibrary.dccs.maya.core import factory
            return type.__call__(factory.MayaItemsFactory, *args, **kwargs)
        else:
            return type.__call__(BaseItemsFactory, *args, **kwargs)


class BaseItemsFactory(plugin.PluginFactory):
    def __init__(self, paths=None):
        super(BaseItemsFactory, self).__init__(
            interface=item.ItemView, plugin_id='NAME', version_id='VERSION', paths=paths)

        self._register_default_paths()

        self._potential_views = sorted(self.plugins(), key=lambda p: p.PRIORITY)
        self._default_view = self.get_plugin_from_id('Data View')

    def _register_default_paths(self):
        self.register_path(
            os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'data'))

    def get_view(self, data_instance):
        """
        Returns the view to be used by the given DataPart
        :param data_instance: DataPart instance
        :return: DataItemView
        """

        for item_view in self._potential_views:
            for component in data_instance.components():
                if str(component) in item_view.REPRESENTING:
                    return item_view

        return self._default_view

    def get_show_save_widget_function(self, data_instance):
        if data_instance == folder.FolderData:
            return folder_view.FolderItemView.show_save_widget

        return base.BaseDataItemView.show_save_widget

    def get_save_widget_class(self, data_instance):
        return save.SaveWidget

    def get_load_widget_class(self, data_instance):
        return load.LoadWidget

    def get_export_widget_class(self, data_instance):
        return export.ExportWidget


@decorators.add_metaclass(_MetaItemsFactory)
class ItemsFactory(object):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains tpDcc-tools-datalibrary server implementation for Maya
"""

from __future__ import print_function, division, absolute_import

import os

import maya.cmds

from tpDcc.core import server
from tpDcc.libs.python import path as path_utils

from tpDcc.libs.datalibrary.core import datalib


class DataLibraryServer(server.DccServer, object):

    PORT = 28231

    def __init__(self, *args, **kwargs):
        super(DataLibraryServer, self).__init__(*args, **kwargs)

        self._data_library = None

    def load_data_items(self, data, reply):

        from tpDcc.libs import datalibrary

        paths_to_load = list()

        dcc_items_path = os.path.join(datalibrary.__path__[0], 'dccs', self.dcc.get_name(), 'data')
        if os.path.isdir(dcc_items_path):
            paths_to_load.append(dcc_items_path)

        reply['result'] = paths_to_load
        reply['success'] = True

    def list_namespaces(self, data, reply):

        exclude_list = ['UI', 'shared']

        namespaces = maya.cmds.namespaceInfo(listOnlyNamespaces=True, recurse=True)
        namespaces = list(set(namespaces) - set(exclude_list))
        namespaces = sorted(namespaces)

        reply['result'] = namespaces
        reply['success'] = True

    def list_nodes(self, data, reply):

        node_name = data.get('node_name', None)
        node_type = data.get('node_type', None)
        full_path = data.get('full_path', True)

        nodes = None
        if not node_name and not node_type:
            nodes = maya.cmds.ls(long=full_path)
        else:
            if node_name and node_type:
                nodes = maya.cmds.ls(node_name, type=node_type, long=full_path)
            elif node_name and not node_type:
                nodes = maya.cmds.ls(node_name, long=full_path)
            elif not node_name and node_type:
                nodes = maya.cmds.ls(type=node_type, long=full_path)

        reply['result'] = nodes or list()
        reply['success'] = True

    def set_focus(self, data, reply):

        ui_name = data.get('ui_name', None)
        try:
            maya.cmds.setFocus(ui_name)
        except Exception:
            pass

        reply['success'] = True

    def save_data(self, data, reply):
        library_path = data['library_path']
        data_path = data['data_path']
        values = data['values']

        if not library_path or not os.path.isfile(library_path):
            reply['success'] = False
            reply['msg'] = 'Impossible to save data "{}" because library path "{}" does not exist!'.format(
                data_path, library_path)
            return

        data_lib = self._get_data_library(library_path)
        data_item = data_lib.get(data_path, only_extension=True)
        if not data_item:
            reply['success'] = False
            reply['msg'] = 'Impossible to retrieve data "{}" from data library: "{}"!'.format(data_path, data_lib)
            return

        save_function = data_item.functionality().get('save')
        if not save_function:
            reply['success'] = False
            reply['msg'] = 'Save functionality is not available for data: "{}"'.format(data_item)
            return

        result = save_function(**values)

        reply['success'] = True
        reply['result'] = result

    def export_data(self, data, reply):
        library_path = data['library_path']
        data_path = data['data_path']
        values = data['values']

        if not library_path or not os.path.isfile(library_path):
            reply['success'] = False
            reply['message'] = 'Impossible to export data "{}" because library path "{}" does not exist!'.format(
                data_path, library_path)
            return

        data_lib = self._get_data_library(library_path)
        data_item = data_lib.get(data_path, only_extension=True)
        if not data_item:
            reply['success'] = False
            reply['message'] = 'Impossible to retrieve data "{}" from data library: "{}"!'.format(data_path, data_lib)
            return

        export_function = data_item.functionality().get('save')
        if not export_function:
            reply['success'] = False
            reply['message'] = 'Export functionality is not available for data: "{}"'.format(data_item)
            return

        result = export_function(**values)

        reply['success'] = True
        reply['result'] = result

    def load_data(self, data, reply):
        library_path = data['library_path']
        data_path = data['data_path']

        if not library_path or not os.path.isfile(library_path):
            reply['success'] = False
            reply['message'] = 'Impossible to load data "{}" because library path "{}" does not exist!'.format(
                data_path, library_path)
            return

        if not os.path.isfile(data_path):
            reply['success'] = False
            reply['message'] = 'Impossible to load data "{}" because it does not exist!'.format(data_path)
            return

        data_lib = self._get_data_library(library_path)
        data_item = data_lib.get(data_path)
        if not data_item:
            reply['success'] = False
            reply['message'] = 'Impossible to retrieve data "{}" from data library: "{}"!'.format(data_path, data_lib)
            return

        load_function = data_item.functionality().get('load')
        if not load_function:
            reply['success'] = False
            reply['message'] = 'Load functionality is not available for data: "{}"'.format(data_item)
            return

        load_function()

        reply['success'] = True

    def import_data(self, data, reply):
        library_path = data['library_path']
        data_path = data['data_path']

        if not library_path or not os.path.isfile(library_path):
            reply['success'] = False
            reply['message'] = 'Impossible to import data "{}" because library path "{}" does not exist!'.format(
                data_path, library_path)
            return

        if not os.path.isfile(data_path):
            reply['success'] = False
            reply['message'] = 'Impossible to import data "{}" because it does not exist!'.format(data_path)
            return

        data_lib = self._get_data_library(library_path)
        data_item = data_lib.get(data_path)
        if not data_item:
            reply['success'] = False
            reply['message'] = 'Impossible to retrieve data "{}" from data library: "{}"!'.format(data_path, data_lib)
            return

        import_function = data_item.functionality().get('import_data')
        if not import_function:
            reply['success'] = False
            reply['message'] = 'Import functionality is not available for data: "{}"'.format(data_item)
            return

        import_function()

        reply['success'] = True

    def reference_data(self, data, reply):
        library_path = data['library_path']
        data_path = data['data_path']

        if not library_path or not os.path.isfile(library_path):
            reply['success'] = False
            reply['message'] = 'Impossible to reference data "{}" because library path "{}" does not exist!'.format(
                data_path, library_path)
            return

        if not os.path.isfile(data_path):
            reply['success'] = False
            reply['message'] = 'Impossible to reference data "{}" because it does not exist!'.format(data_path)
            return

        data_lib = self._get_data_library(library_path)
        data_item = data_lib.get(data_path)
        if not data_item:
            reply['success'] = False
            reply['message'] = 'Impossible to retrieve data "{}" from data library: "{}"!'.format(data_path, data_lib)
            return

        reference_function = data_item.functionality().get('reference_data')
        if not reference_function:
            reply['success'] = False
            reply['message'] = 'Reference functionality is not available for data: "{}"'.format(data_item)
            return

        reference_function()

        reply['success'] = True

    def _get_data_library(self, library_path):
        if self._data_library and path_utils.clean_path(
                self._data_library.identifier) == path_utils.clean_path(library_path):
            return self._data_library

        self._data_library = datalib.DataLibrary.load(library_path)

        return self._data_library



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Initialization module for tpDcc-tools-datalibrary
"""

from __future__ import print_function, division, absolute_import

import os
import logging.config


def create_logger(dev=False):
    """
    Creates logger for current tpDcc-tools-datalibrary package
    """

    logger_directory = os.path.normpath(os.path.join(os.path.expanduser('~'), 'tpDcc', 'logs', 'tools'))
    if not os.path.isdir(logger_directory):
        os.makedirs(logger_directory)

    logging_config = os.path.normpath(os.path.join(os.path.dirname(__file__), '__logging__.ini'))

    logging.config.fileConfig(logging_config, disable_existing_loggers=False)
    logger = logging.getLogger('tpDcc-tools-datalibrary')
    dev = os.getenv('TPDCC_DEV', dev)
    if dev:
        logger.setLevel(logging.DEBUG)
        for handler in logger.handlers:
            handler.setLevel(logging.DEBUG)

    return logger


create_logger()


if __name__ == '__main__':
    import tpDcc.loader
    from tpDcc.managers import tools

    tpDcc.loader.init(dev=False)

    tools.ToolsManager().launch_tool_by_id('tpDcc-tools-datalibrary')



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains utils functions to handle persistent wigdet values settings
"""

from __future__ import print_function, division, absolute_import

import os

from tpDcc.managers import libs
from tpDcc.libs.python import path as path_utils

from tpDcc.tools.datalibrary.core import utils


_SETTINGS = dict()


def path():
    """
    Returns settings path
    :return: str
    """

    libs_file_path = libs.LibsManager().get_library_settings_file_path('tpDcc-libs-datalibrary')

    return path_utils.join_path(os.path.dirname(libs_file_path), 'widgetSettings.json')


def read():
    """
    Returns widget related settings
    :return: dict
    """

    global _SETTINGS
    if not _SETTINGS:
        _SETTINGS = utils.read_json(path())

    return _SETTINGS


def save(data):
    """
    Saves the given data dictionary to settings
    :param data: dict
    """

    global _SETTINGS
    _SETTINGS = dict()
    utils.update_json(path(), data)


def get(key, default=None):
    """
    Returns value from disk
    :param key: str
    :param default: object
    :return: object
    """

    return read().get(key, default)


def set(key, value):
    """
    Sets settings value to disk
    :param key: str
    :param value: object
    """

    save({key: value})


def reset():
    """
    Removes and resets the item settings
    """

    global _SETTINGS
    _SETTINGS = dict()
    utils.remove_path(path())



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains data library window widget implementation
"""

from __future__ import print_function, division, absolute_import

import os
import re
import time
import copy
import logging
import operator
from functools import partial

from Qt.QtCore import Qt, Signal, QPoint
from Qt.QtWidgets import QSizePolicy, QWidget, QFrame, QSplitter, QFileDialog, QDialogButtonBox, QMenu, QAction
from Qt.QtGui import QCursor, QColor, QIcon, QKeyEvent, QStatusTipEvent

from tpDcc.managers import configs, resources
from tpDcc.libs.python import python, osplatform, fileio, folder, path as path_utils
from tpDcc.libs.resources.core import icon
from tpDcc.libs.qt.core import qtutils, base, animation, menu, decorators as qt_decorators
from tpDcc.libs.qt.widgets import layouts, stack, toolbar, messagebox
from tpDcc.libs.datalibrary.core import datalib

from tpDcc.tools.datalibrary.core import consts, utils, factory
from tpDcc.tools.datalibrary.core.views import item as items_view
from tpDcc.tools.datalibrary.widgets import viewer, search, sidebar, status
from tpDcc.tools.datalibrary.widgets.menus import filter, group, sort, libraries

logger = logging.getLogger(consts.TOOL_ID)


class PreviewFrame(QFrame):
    pass


class SidebarFrame(QFrame):
    pass


class LibraryWindow(base.BaseWidget):

    DEFAULT_NAME = consts.DEFAULT_LIBRARY_NAME

    DEFAULT_SETTINGS = {
        "library": {
            "sort_by": ["name:asc"],
            "group_by": ["type:asc"]
        },
        "paneSizes": [130, 280, 180],
        "trashFolderVisible": False,
        "sidebarWidgetVisible": True,
        "previewWidgetVisible": True,
        "toolBarWidgetVisible": True,
        "statusBarWidgetVisible": True,
        "recursiveSearchEnabled": True,
        "viewerWidget": {
            "spacing": 2,
            "padding": 6,
            "zoomAmount": 80,
            "textVisible": True,
        },
        "searchWidget": {
            "text": "",
        },
        "filterByMenu": {
            "Folder": False
        }
    }

    PROGRESS_BAR_VISIBLE = consts.PROGRESS_BAR_VISIBLE
    TRASH_ENABLED = True
    TEMP_PATH_MENU_ENABLED = False
    DPI_ENABLED = False

    LIBRARY_CLASS = datalib.DataLibrary
    VIEWER_CLASS = viewer.DataViewer
    SEARCH_WIDGET_CLASS = search.DataSearcherWidget
    STATUS_WIDGET_CLASS = status.DataStatusWidget
    TOOLBAR_WIDGET_CLASS = toolbar.ToolBar
    SIDEBAR_WIDGET_CLASS = sidebar.SidebarWidget
    SORTBY_MENU_CLASS = sort.SortByMenu
    GROUPBY_MENU_CLASS = group.GroupByMenu
    FILTERBY_MENU_CLASS = filter.FilterByMenu

    itemSelectionChanged = Signal(object)
    folderSelectionChanged = Signal(object)

    def __init__(
            self, settings, name='', items_factory=None,
            json_settings_file_path='', library_path='', client=None, parent=None):

        self._dpi = 1.0
        self._client = client
        self._settings = settings
        self._name = name or self.DEFAULT_NAME
        self._items = list()
        self._is_locked = False
        self._is_loaded = False
        self._library = None
        self._refresh_enabled = False
        self._trash_enabled = self.TRASH_ENABLED
        self._superusers = None
        self._lock_reg_exp = None
        self._unlock_reg_exp = None
        self._kwargs = dict()
        self._path = None
        self._menu_items = list()
        self._repository_type = None
        self._repository_path = ''

        self._preview_widget = None
        self._new_item_widget = None
        self._current_item = None
        self._items_factory = items_factory

        self._items_hidden_count = 0
        self._items_visible_count = 0

        self._is_trash_folder_visible = False
        self._sidebar_widget_visible = True
        self._preview_widget_visible = True
        self._status_widget_visible = True

        self._libraries_menu = None
        self._settings_file_path = json_settings_file_path or utils.settings_path()

        super(LibraryWindow, self).__init__(parent)

        # # TODO: THIS IS FOR DEV
        # library_path = r'D:\rigs\rigscript\chimp\data.db'
        self.set_library(library_path or self._settings.get('last_path'))

        self.set_dpi(1.0)

    # ============================================================================================================
    # PROPERTIES
    # ============================================================================================================

    @property
    def client(self):
        return None if not self._client else self._client

    @property
    def factory(self):
        return self._items_factory

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def get_main_layout(self):
        return layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))

    def ui(self):
        super(LibraryWindow, self).ui()

        self.setMinimumWidth(5)
        self.setMinimumHeight(5)

        self._stack = stack.SlidingStackedWidget()
        self.main_layout.addWidget(self._stack)

        self._sidebar_frame = SidebarFrame(self)
        sidebar_frame_lyt = layouts.VerticalLayout(margins=(0, 1, 0, 0))
        self._sidebar_frame.setLayout(sidebar_frame_lyt)

        self._new_item_frame = PreviewFrame(self)
        self._new_item_frame.setMinimumWidth(5)
        new_item_lyt = layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))
        self._new_item_frame.setLayout(new_item_lyt)

        self._preview_frame = PreviewFrame(self)
        self._preview_frame.setMinimumWidth(5)
        preview_frame_lyt = layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))
        self._preview_frame.setLayout(preview_frame_lyt)

        self._viewer = self.VIEWER_CLASS(self)
        self._viewer.installEventFilter(self)
        self._viewer.setContextMenuPolicy(Qt.CustomContextMenu)

        self._toolbar_widget = self.TOOLBAR_WIDGET_CLASS(self)
        self._search_widget = self.SEARCH_WIDGET_CLASS(self)
        self._sort_by_menu = self.SORTBY_MENU_CLASS(self)
        self._group_by_menu = self.GROUPBY_MENU_CLASS(self)
        self._filter_by_menu = self.FILTERBY_MENU_CLASS(self)
        self._status_widget = self.STATUS_WIDGET_CLASS(self)
        self._sidebar_widget = self.SIDEBAR_WIDGET_CLASS(self)
        self._sidebar_widget.setContextMenuPolicy(Qt.CustomContextMenu)
        sidebar_frame_lyt.addWidget(self._sidebar_widget)

        self._splitter = QSplitter(Qt.Horizontal, self)
        self._splitter.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Expanding)
        self._splitter.setHandleWidth(2)
        self._splitter.setChildrenCollapsible(False)
        self._splitter.addWidget(self._sidebar_frame)
        self._splitter.addWidget(self._viewer)
        self._splitter.addWidget(self._preview_frame)
        self._splitter.setStretchFactor(0, False)
        self._splitter.setStretchFactor(1, True)
        self._splitter.setStretchFactor(2, False)

        base_widget = QWidget()
        base_layout = layouts.VerticalLayout(spacing=0, margins=(0, 0, 0, 0))
        base_widget.setLayout(base_layout)
        base_layout.addWidget(self._toolbar_widget)
        base_layout.addWidget(self._splitter)
        base_layout.addWidget(self._status_widget)

        self._stack.addWidget(base_widget)
        self._stack.addWidget(self._new_item_frame)

        self._setup_toolbar()

    def setup_signals(self):
        self.folderSelectionChanged.connect(self.update_lock)
        self._sidebar_widget.itemSelectionChanged.connect(self._on_folder_selection_changed)
        self._sidebar_widget.itemDropped.connect(self._on_item_dropped)
        self._sidebar_widget.customContextMenuRequested.connect(self._on_show_folders_menu)
        self._sidebar_widget.settingsMenuRequested.connect(self._on_sidebar_settings_menu_requested)
        self._viewer.itemSelectionChanged.connect(self._on_item_selection_changed)
        self._viewer.itemMoved.connect(self._on_item_moved)
        self._viewer.itemDropped.connect(self._on_item_dropped)
        self._viewer.keyPressed.connect(self._on_key_pressed)
        self._viewer.customContextMenuRequested.connect(self._on_show_items_context_menu)

    def event(self, event):
        """
        Overrides LibraryWindow event function
        """

        if isinstance(event, QKeyEvent):
            if qtutils.is_control_modifier() and event.key() == Qt.Key_F:
                self.search_widget().setFocus()
        elif isinstance(event, QStatusTipEvent):
            self.status_widget().show_info_message(event.tip())

        return super(LibraryWindow, self).event(event)

    def showEvent(self, event):
        """
        Overrides LibraryWindow showEvent function
        """

        super(LibraryWindow, self).showEvent(event)

        if not self.is_loaded():
            self._is_loaded = True
            self.load_settings()
            self.set_refresh_enabled(True)

    # ============================================================================================================
    # CLASS METHODS
    # ============================================================================================================

    @classmethod
    def create_action(cls, item_class, menu, library_window):
        """
        Returns the action to be displayed when the user clicks the "Add New Item" icon
        :param item_class: str
        :param menu: QMenu
        :param library_window: LibraryWindow
        :return: QAction
        """

        if not item_class.MENU_NAME:
            return

        show_save_widget = library_window.factory.get_show_save_widget_function(item_class)

        icon_name = os.path.splitext(item_class.MENU_ICON)[0] if item_class.MENU_ICON else None
        icon_name = icon_name or 'tpDcc'
        action_icon = resources.icon(icon_name) if icon_name else QIcon()
        if action_icon.isNull():
            action_icon = resources.icon('tpDcc')
        callback = partial(show_save_widget, item_class, library_window)
        action = QAction(action_icon, item_class.MENU_NAME, menu)
        action.triggered.connect(callback)
        menu.addAction(action)

        return action

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def name(self):
        """
        Returns library name
        :return: str
        """

        return self._name

    def library(self):
        """
        Returns data library being used
        :return: DataLibrary
        """

        return self._library

    def set_library(self, library):
        """
        Sets data library to use
        :param library: str or DataLibrary
        """

        if self._library and library and self._library == library:
            return

        if not library and self._library:
            try:
                qtutils.safe_disconnect_signal(self._library.dataChanged)
            except Exception as exc:
                pass

        if python.is_string(library):
            if not self._library or library != self._library.identifier:
                if not library:
                    logger.warning('Given library path "{}" does not exists!'.format(library))
                else:
                    if not os.path.isfile(library):
                        self._library = self.LIBRARY_CLASS.create(library)
                    else:
                        self._library = self.LIBRARY_CLASS.load(library)
                    self._library.add_scan_location(path_utils.clean_path(os.path.join(os.path.dirname(library))))
        else:
            self._library = library

        if self._library:
            plugin_locations = self._library.plugin_locations() or list()

            # Create factory to hold all available item views
            if not self._items_factory:
                self._items_factory = factory.ItemsFactory(paths=plugin_locations)

            self._path = path_utils.clean_path(os.path.dirname(self.database_path()))

            # This is very time consuming, we should avoid calling this
            self._library.sync()

            # Add some default queries
            self._library.add_query(
                {'name': 'invalid extensions', 'operator': 'and', 'filters': [('extension', 'not', '.pyc')]}
            )

            self._library.dataChanged.connect(self.refresh)

        self._sort_by_menu.set_library(self._library)
        self._group_by_menu.set_library(self._library)
        self._filter_by_menu.set_library(self._library)
        self._viewer.set_library(self._library)
        self._search_widget.set_library(self._library)
        self._sidebar_widget.set_library(self._library)

        self.set_refresh_enabled(True)
        self.update_view_button()
        self.update_filters_button()
        self.update_preview_widget()

    def database_path(self):
        """
        Returns path to the database showed by the window
        :return: str
        """

        if not self._library:
            return

        return self._library.identifier

    def path(self):
        """
        Returns the path being used by the library
        :return: str
        """

        return self._path

    def set_path(self, path):
        path = path_utils.clean_path(path)

        if path == self.path():
            return

        self._path = path

        if not self._library:
            self.set_library(self._path)

        if not os.path.exists(self.database_path()):
            self.sync()

        self.refresh()
        self.library().search()
        self.update_preview_widget()

    def is_loaded(self):
        """
        Returns whether or not Data library window has been loaded
        :return: bool
        """

        return self._is_loaded

    def is_refresh_enabled(self):
        """
        Returns whether refresh is enabled or not
        If not, all updates will be ignored
        :return: bool
        """

        return self._refresh_enabled

    def set_refresh_enabled(self, flag):
        """
        Whether widgets should be updated or not
        :param flag: bool
        """

        library = self.library()
        if not library:
            return

        library.set_search_enabled(flag)
        self._refresh_enabled = flag

    def is_recursive_search_enabled(self):
        """
        Returns whether recursive search is enabled or not
        :return: bool
        """

        return self.sidebar_widget().is_recursive()

    def set_recursive_search_enabled(self, flag):
        """
        Sets whether recursive search is enabled or not
        :param flag: bool
        """

        self.sidebar_widget().set_recursive(flag)

    def clean_library(self):
        """
        Cleans library data by removing files that are not being managed by database
        NOTE: This is a destructive operation
        :return:
        """

        if not self.library():
            return

        # TODO: This operation should only be available for superusers

        res = messagebox.MessageBox.question(
            self, 'Clean Library', 'This is a destructive operation. Files and folders not being managed by database '
                                   'will be removed. Is recommended to backup your data before proceeding with this '
                                   'operation. Are you sure you want to continue?')
        if res != QDialogButtonBox.Yes:
            return

        self.library().cleanup()

    def superusers(self):
        """
        Returns the superusers for the window
        :return: list(str)
        """

        return self._superusers

    def set_superusers(self, superusers):
        """
        Sets the valid superusres for the library widget
        This will lock all folders unless the current user is a superuser
        :param superusers: list(str)
        """

        self._superusers = superusers
        self.update_lock()

    def lock_reg_exp(self):
        """
        Returns the lock regular expression used for locking the widget
        :return: str
        """

        return self._lock_reg_exp

    def set_lock_reg_exp(self, reg_exp):
        """
        Sets the lock regular expression used for locking the widget
        Locks only folders that contain the given regular expression in their path
        :param reg_exp: str
        """

        self._lock_reg_exp = reg_exp
        self.update_lock()

    def unlock_reg_exp(self):
        """
        Returns the unlock regular expression used for unlocking the widget
        :return: str
        """

        return self._unlock_reg_exp

    def set_unlock_reg_exp(self, reg_exp):
        """
        Sets the unlock regular expression used for locking the widget
        Unlocks only folders that contain the given regular expression in their path
        :param reg_exp: str
        """

        self._unlock_reg_exp = reg_exp
        self.update_lock()

    def is_lock_reg_exp_enabled(self):
        """
        Returns whether the lock regular expression or unlock regular expression has been set
        :return: bool
        """

        return not (self.superusers() is None and self.lock_reg_exp() is None and self.unlock_reg_exp() is None)

    def refresh(self):
        """
        Refresh all necessary items
        """

        if self.is_refresh_enabled():
            self.update()

    def update(self):
        """
        Overrides base QMainWindow update function
        Update the library widget and the data
        """

        self.refresh_sidebar()
        self.update_window_title()

    def update_window_title(self):
        """
        Updates the window title
        """

        window_title = self.windowTitle()

        if self.is_locked():
            window_title += ' (Locked)'

        self.setWindowTitle(window_title)

    def update_filters_button(self):
        """
        Updates the icon for the filters menu
        """

        action = self.toolbar_widget().find_action('Filters')
        filter_icon = resources.icon('filter', theme='black')
        if filter_icon.isNull():
            return
        # icon.set_color(self.icon_color())
        filter_icon.set_color(QColor(255, 255, 255, 255))
        if self._filter_by_menu.is_active():
            filter_icon.set_badge(18, 1, 9, 9, color=consts.ICON_BADGE_COLOR)
        action.setIcon(filter_icon)

    def update_lock(self):
        """
        Updates the lock state for the library
        """

        if not self.is_lock_reg_exp_enabled():
            return

        superusers = self.superusers() or list()
        re_locked = re.compile(self.lock_reg_exp() or '')
        re_unlocked = re.compile(self.unlock_reg_exp() or '')

        if osplatform.get_user(lower=True) in superusers:
            self.set_locked(False)
        elif re_locked.match('') and re_unlocked.match(''):
            if superusers:
                self.set_locked(True)
            else:
                self.set_locked(False)
        else:
            # Lock/Unlock folders matching regex
            folders = self.selected_folder_paths()
            if not re_locked.match(''):
                for folder_found in folders:
                    if re_locked.search(folder_found):
                        self.set_locked(True)
                        return
                self.set_locked(False)
            if not re_unlocked.match(''):
                for folder_found in folders:
                    if re_unlocked.search(folder_found):
                        self.set_locked(False)
                        return

    @qt_decorators.show_wait_cursor
    def sync(self, force_start=True, force_search=True):
        """
        Sync any data that might be out of date with the model
        """

        def _sync():
            start_time = time.time()
            self.library().sync(progress_callback=self._set_progress_bar_value)
            elapsed_time = time.time() - start_time
            self.status_widget().show_info_message('Synced items in {0:.3f} seconds'.format(elapsed_time))
            self._set_progress_bar_value('Done')
            if force_start:
                progress_bar.close()
            else:
                animation.fade_out_widget(progress_bar, duration=500, on_finished=progress_bar.close)

            if force_search and self.library():
                # self.library().set_dirty(True)
                self.library().search()

        progress_bar = self.status_widget().progress_bar()
        self._set_progress_bar_value('Syncing')
        if self.PROGRESS_BAR_VISIBLE:
            progress_bar.show()

        if force_start:
            _sync()
        else:
            animation.fade_in_widget(progress_bar, duration=1, on_finished=_sync)

    def set_sizes(self, sizes):
        """
        :type sizes: (int, int, int)
        :rtype: None
        """
        first_size, second_size, third_size = sizes

        if third_size == 0:
            third_size = 200

        if first_size == 0:
            first_size = 120

        self._splitter.setSizes([first_size, second_size, third_size])
        self._splitter.setStretchFactor(1, 1)

    # ============================================================================================================
    # DPI
    # ============================================================================================================

    def is_dpi_enabled(self):
        """
        Returns whether or not DPI functionality is enabled
        :return: bool
        """

        dpi_enabled = self.DPI_ENABLED
        datalib_config = configs.get_library_config('tpDcc-tools-datalibrary')
        if datalib_config:
            dpi_enabled = datalib_config.get('dpi_enabled') or dpi_enabled

        return dpi_enabled

    def dpi(self):
        """
        Returns current library wiget DPI
        :return: float
        """

        if not self.is_dpi_enabled():
            return 1.0

        return float(self._dpi)

    def set_dpi(self, dpi):
        """
        Sets the current DPI for library widget
        :param dpi: float
        """

        if not self.is_dpi_enabled():
            dpi = 1.0

        self._dpi = dpi

        self.viewer().set_dpi(dpi)
        self.toolbar_widget().set_dpi(dpi)
        self.sidebar_widget().set_dpi(dpi)
        self.status_widget().setFixedHeight(20 * dpi)

        self._splitter.setHandleWidth(2 * dpi)

        self.show_toast_message('DPI: {}'.format(int(dpi * 100)))

        # self.reload_stylesheet()

    # ============================================================================================================
    # VERSION
    # ============================================================================================================

    def get_repository_type(self):
        if self._repository_type:
            return self._repository_type

        if not self._settings:
            return None

        return self._settings.get('version_control', default_value=0, setting_group='Version', begin_group='Repository')

    def set_repository_type(self, repository_type_index):
        try:
            repository_type_index = int(repository_type_index)
        except Exception:
            return
        self._repository_type = repository_type_index

    def get_repository_path(self):
        if not self._library:
            return None

        library_path = self._library.identifier
        if not os.path.isfile(library_path):
            return None

        return os.path.dirname(library_path)

    # ============================================================================================================
    # LOCK / UNLOCK
    # ============================================================================================================

    def is_locked(self):
        """
        Returns whether the library is locked or not
        :return: bool
        """

        return self._is_locked

    def set_locked(self, flag):
        """
        Sets the state of the library to be editable or not
        :param flag: bool
        """

        self._is_locked = flag

        self.sidebar_widget().set_locked(flag)
        self.viewer().set_locked(flag)
        self.update_create_item_button()
        self.update_window_title()
        self.lockChanged.emit(flag)

    def update_create_item_button(self):
        """
        Updates the create item icon depending of the lock status
        """

        action = self.toolbar_widget().find_action('New Item')
        if self.is_locked():
            icon = resources.icon('lock')
            action.setEnabled(False)
        else:
            icon = resources.icon('plus')
            action.setEnabled(True)
        # icon.set_color(self.icon_color())
        action.setIcon(icon)

    # ============================================================================================================
    # TOOLBAR WIDGET
    # ============================================================================================================

    def toolbar_widget(self):
        """
        Returns menu bar widget
        :return: toolbar.ToolBar
        """

        return self._toolbar_widget

    def add_toolbar_action(self, name, icon, tip, callback=None):
        """
        Internal function that adds a button/action into menu bar widget
        :param name: str
        :param icon: QIcon
        :param tip: str
        :param callback: fn
        :return: QAction
        """

        # We need to do this to avoid PySide2 errors
        def _callback():
            callback()

        action = self.toolbar_widget().addAction(name)
        if icon:
            action.setIcon(icon)
        if tip:
            action.setToolTip(tip)
            action.setStatusTip(tip)
        if callback:
            action.triggered.connect(_callback)

        return action

    def is_toolbar_widget_visible(self):
        """
        Returns whether toolbar widget is visible or not
        :return: bool
        """

        return self.toolbar_widget().is_expanded()

    def set_toolbar_widget_visible(self, flag):
        """
        Sets whether menubar widget is visible or not
        :param flag: bool
        """

        flag = bool(flag)
        if flag:
            self.toolbar_widget().expand()
        else:
            self.toolbar_widget().collapse()

    # ============================================================================================================
    # SEARCH WIDGET
    # ============================================================================================================

    def search_widget(self):
        """
        Returns search widget
        :return: LibrarySearchWidget
        """

        return self._search_widget

    def set_search_text(self, text):
        """
        Set the search widget text
        :param text: str
        """

        self.search_widget().setText(text)

    def items_visible_count(self):
        """
        Return the number of visible items
        :return: int
        """

        return self._items_visible_count

    def items_hidden_count(self):
        """
        Return the number of hidden items
        :return: int
        """

        return self._items_hidden_count

    # ============================================================================================================
    # SIDEBAR WIDGET
    # ============================================================================================================

    def sidebar_widget(self):
        """
        Return the sidebar widget
        :return: LibrarySidebarWidget
        """

        return self._sidebar_widget

    def is_sidebar_widget_visible(self):
        """
        Return whether SideBar widget is visible or not
        :return: bool
        """

        return self._sidebar_widget_visible

    def set_sidebar_widget_visible(self, flag):
        """
        Set whether SideBar widget is visible or not
        :param flag: bool
        """

        flag = bool(flag)
        self._sidebar_widget_visible = flag

        if flag:
            self._sidebar_frame.setVisible(True)
        else:
            self._sidebar_frame.setVisible(False)

        self.update_view_button()

    def set_folder_data(self, path, data):
        """
        Sets sidebar folder data
        :param path: str
        :param data: dict
        """

        self.sidebar_widget().set_item_data(path, data)

    @qt_decorators.show_wait_cursor
    def refresh_sidebar(self):
        """
        Refresh the state of the sidebar widget
        """

        path = self.path()
        if not path and self._allow_non_path:
            return
        else:
            if not path:
                return self.show_hello_dialog()
            elif not os.path.exists(path):
                return self.show_path_error_dialog()

            self.update_sidebar()

    def update_sidebar(self):
        """
        Update the folders to be shown in the folders widget
        """

        library = self.library()
        if not library:
            return

        current_data = dict()
        root = self.path()

        root_identifier = self.library().get_identifier(root)
        queries = [{'operator': 'and',
                    'filters': [('folder', 'is', 'True'), ('directory', 'startswith', root_identifier)]}]

        items = self.library().find_items(queries)
        for item in items:
            if not item:
                continue
            current_data[item.format_identifier()] = item.data()

        self.sidebar_widget().set_data(current_data, root=root)

    def selected_folder_path(self):
        """
        Return the selected folder items
        :return: str or None
        """

        return self.sidebar_widget().selected_path()

    def selected_folder_paths(self):
        """
        Return the selected folder items
        :return: list(str)
        """

        return self.sidebar_widget().selected_paths()

    def select_folder_path(self, path):
        """
        Select the given folder path
        :param path: str
        """

        self.select_folder_paths([path])

    def select_folder_paths(self, paths):
        """
        Select the given folder paths
        :param paths: list(str)
        """

        self.sidebar_widget().select_paths(paths)

    def create_folder_context_menu(self):
        """
        Returns the folder menu for the selected folders
        :return: QMenu
        """

        path = self.selected_folder_path()
        items = list()
        if path:
            queries = [{"filters": [("path", "is", path)]}]
            items = self.library().find_items(queries)
            if items:
                self._items = items

        return self._create_item_context_menu(items)

    # ============================================================================================================
    # VIEWER WIDGET
    # ============================================================================================================

    def viewer(self):
        """
        Returns muscle viewer widget
        :return:
        """

        return self._viewer

    def items(self):
        """
        Return all the loaded items
        :return: list(LibraryItem)
        """

        return self._items

    def selected_items(self):
        """
        Return selected items
        :return: list(LibraryItem)
        """

        return self.viewer().selected_items()

    # ============================================================================================================
    # STATUS WIDGET
    # ============================================================================================================

    def status_widget(self):
        """
        Returns the status widget
        :return: StatusWidget
        """

        return self._status_widget

    def is_status_widget_visible(self):
        """
        Return whether StatusWidget is visible or not
        :return: bool
        """

        return self._status_widget_visible

    def set_status_widget_visible(self, flag):
        """
        Set whether StatusWidget is visible or not
        :param flag: bool
        """

        flag = bool(flag)
        self._status_widget_visible = flag
        self.status_widget().setVisible(flag)

    # ============================================================================================================
    # PREVIEW WIDGET
    # ============================================================================================================

    def preview_widget(self):
        """
        Returns the current preview widget
        :return: QWidget
        """

        return self._preview_widget

    def set_preview_widget(self, widget):
        """
        Set the preview widget
        :param widget: QWidget
        """

        if self._preview_widget == widget:
            msg = 'Preview widget already contains widget {}'.format(widget)
            logger.debug(msg)
        else:
            self.close_preview_widget()
            self._preview_widget = widget
            if self._preview_widget:
                self._preview_frame.layout().addWidget(self._preview_widget)
                self._preview_widget.show()

    def is_preview_widget_visible(self):
        """
        Returns whether preview widget is visible or not
        :return: bool
        """

        return self._preview_widget_visible

    def set_preview_widget_visible(self, flag):
        """
        Set if the PreviewWidget should be showed or not
        :param flag: bool
        """

        flag = bool(flag)
        self._preview_widget_visible = flag

        if flag:
            self._preview_frame.setVisible(True)
        else:
            self._preview_frame.setVisible(False)

        self.update_view_button()

    def update_preview_widget(self):
        """
        Update the current preview widget
        """

        self.set_preview_widget_from_item(self._current_item, force=True)

    def set_preview_widget_from_item(self, item, force=True):
        """
        Set the preview widget from the given item
        :param item: LibvraryItem
        :param force: bool
        """

        if not force and self._current_item == item:
            logger.debug('The current item preview widget is already set!')
            return

        self._current_item = item

        if item:
            self.close_preview_widget()
            try:
                item.show_preview_widget(self)
            except Exception as e:
                self.show_error_message(e)
                self.clear_preview_widget()
                raise
        else:
            self.clear_preview_widget()

    def clear_preview_widget(self):
        """
        Set the default preview widget
        """

        self._preview_widget = None
        widget = base.PlaceholderWidget()
        self.set_preview_widget(widget)

    def close_preview_widget(self):
        """
        Close and delete the preview widget
        """

        frame_layout = self._preview_frame.layout()

        while frame_layout.count():
            item = frame_layout.takeAt(0)
            item.widget().hide()
            item.widget().close()
            item.widget().deleteLater()

        self._preview_widget = None

    def set_save_widget(self, create_widget):
        """
        Set the widget that should be showed when saving a new item
        :param create_widget: QWidget
        """

        if not create_widget:
            return

        self.set_new_item_widget_visible(True)

        self.set_new_item_widget(create_widget)
        create_widget.cancelled.connect(partial(self._stack.slide_in_index, 0))
        create_widget.saved.connect(partial(self._stack.slide_in_index, 0))

    def set_export_widget(self, export_widget):
        """
        Sets the widget that should be showed when exporting a new item
        :param export_widget: QWidget
        """

        if not export_widget:
            return

        self.set_new_item_widget_visible(True)

        self.set_new_item_widget(export_widget)
        export_widget.cancelled.connect(partial(self._stack.slide_in_index, 0))
        export_widget.saved.connect(partial(self._stack.slide_in_index, 0))

    # ============================================================================================================
    # NEW ITEM WIDGET
    # ============================================================================================================

    def new_item_widget(self):
        """
        Returns the current new item widget
        :return: QWidget
        """

        return self._new_item_widget

    def set_new_item_widget(self, widget):
        """
        Sets the new item widget
        :param widget: QWidget
        """

        if self._new_item_widget == widget:
            msg = 'New Item widget already contains widget {}'.format(widget)
            logger.debug(msg)
        else:
            self.close_new_item_widget()
            self._new_item_widget = widget
            if self._new_item_widget:
                self._new_item_frame.layout().addWidget(self._new_item_widget)
                self._new_item_widget.show()

    def is_new_item_widget_visible(self):
        """
        Returns whether new item widget is visible or not
        :return: bool
        """

        return self._new_item_widget_visible

    def set_new_item_widget_visible(self, flag):
        """
        Set if the New Item widget should be showed or not
        :param flag: bool
        """

        flag = bool(flag)
        self._new_item_widget_visible = flag

        if flag:
            self._stack.slide_in_index(1)
            self._new_item_frame.show()
        else:
            self._stack.slide_in_index(0)
            self._new_item_frame.hide()

        self.update_view_button()

    def clear_new_item_widget(self):
        """
        Set the default new item widget
        """

        self._new_item_widget = None
        widget = base.PlaceholderWidget()
        self.set_new_item_widget(widget)

    def close_new_item_widget(self):
        """
        Close and delete the new item widget
        """

        lyt = self._new_item_frame.layout()

        while lyt.count():
            item = lyt.takeAt(0)
            item.widget().hide()
            item.widget().close()
            item.widget().deleteLater()

        self._new_item_widget = None

    # ============================================================================================================
    # TRASH FOLDER
    # ============================================================================================================

    def trash_enabled(self):
        """
        Returns True if moving items to trash
        :return: bool
        """

        return self._trash_enabled

    def set_trash_enabled(self, flag):
        """
        Sets if items can be trashed or not
        :param flag: bool
        """

        self._trash_enabled = flag

    def is_path_in_trash(self, path):
        """
        Returns whether given path is in trash or not
        :param path: str
        :return: bool
        """

        return consts.TRASH_NAME in path.lower()

    def trash_path(self):
        """
        Returns the trash path for the library
        :return: str
        """

        path = self.path()
        return '{}/{}'.format(path, consts.TRASH_NAME.title())

    def trash_folder_exists(self):
        """
        Returns whether trash folder exists or not
        :return: bool
        """

        return os.path.exists(self.trash_path())

    def create_trash_folder(self):
        """
        Create the trash folder if it does not already exists
        """

        trash_path = self.trash_path()
        if not os.path.exists(trash_path):
            os.makedirs(trash_path)

    def is_trash_folder_visible(self):
        """
        Return whether trash folder is visible or not
        :return: bool
        """

        return self._is_trash_folder_visible

    def set_trash_folder_visible(self, flag):
        """
        Set whether trash folder is visible or not
        :param flag: bool
        """

        self._is_trash_folder_visible = flag

        if flag:
            query = {
                'name': 'trash_query',
                'filters': list()
            }
        else:
            query = {
                'name': 'trash_query',
                'filters': [('path', 'not_contains', consts.TRASH_NAME.title())]
            }

        # self.library().add_global_query(query)
        # self.update_sidebar()
        # self.library().search()

    def is_trash_selected(self):
        """
        Return whether the selected folders are in the trash or not
        :return: bool
        """

        folders = self.selected_folder_paths()
        for folder in folders:
            if self.is_path_in_trash(folder):
                return True

        items = self.selected_items()
        for item in items:
            if self.is_path_in_trash(item.path()):
                return True

        return False

    def move_items_to_trash(self, items):
        """
        Move the given items to trash path
        :param items: list(LibraryItem)
        """

        self.create_trash_folder()
        self.move_items(items, target_folder=self.trash_path(), force=True)

    def show_move_items_to_trash_dialog(self, items=None):
        """
        Show the "Move to Trash" dialog for the selected items
        :param items: list(LibraryItem) or None
        """

        items = items or self.selected_items()
        if items:
            title = 'Move to Trash?'
            text = 'Are you sure you want to move the selected item/s to the trash?'
            result = self.show_question_dialog(title, text)
            if result == QDialogButtonBox.Yes:
                self.move_items_to_trash(items)

    # ============================================================================================================
    # TOAST/MESSAGES
    # ============================================================================================================

    def show_toast_message(self, text, duration=1000):
        """
        Shows toast widget with the given text and duration
        :param text: str
        :param duration:int
        """

        self.viewer().show_toast_message(text, duration)

    def show_success_message(self, text, msecs=None):
        """
        Shows success message to the user
        :param text: str
        :param msecs: int or None
        """

        self.status_widget().show_ok_message(text, msecs)

    def show_info_message(self, text, msecs=None):
        """
        Shows info message to the user
        :param text: str
        :param msecs: int or None
        """

        self.status_widget().show_info_message(text, msecs)

    def show_warning_message(self, text, msecs=None):
        """
        Shows warning message to the user
        :param text: str
        :param msecs: int or None
        """

        self.status_widget().show_warning_message(text, msecs)
        self.set_status_widget_visible(True)

    def show_error_message(self, text, msecs=None):
        """
        Shows error message to the user
        :param text:str
        :param msecs: int or None
        """

        self.status_widget().show_error_message(text, msecs)
        self.set_status_widget_visible(True)

    def show_refresh_message(self):
        """
        Show long the current refresh took
        """

        item_count = len(self.library().results())
        elapsed_time = self.library().search_time()

        plural = ''
        if item_count > 1:
            plural = 's'

        msg = 'Found {0} item{1} in {2:.3f} seconds.'.format(item_count, plural, elapsed_time)
        self.status_widget().show_info_message(msg)

        logger.debug(msg)

    @qt_decorators.show_arrow_cursor
    def show_hello_dialog(self):
        """
        This function is called when there is not root path set for the library
        """

        text = 'Please choose a folder location for storing the data'
        dialog = messagebox.create_message_box(
            None, 'Welcome {}'.format(self.name()), text, header_pixmap=self._library_icon_path)
        dialog.accepted.connect(self.show_change_path_dialog)
        dialog.show()

    @qt_decorators.show_arrow_cursor
    def show_path_error_dialog(self):
        """
        This function is called when the root path does not exists during refresh
        """

        path = self.path()
        text = 'The current root path does not exists "{}". Please select a new root path to continue.'.format(path)
        dialog = messagebox.create_message_box(self, 'Path Error', text)
        dialog.show()
        dialog.accepted.connect(self.show_change_path_dialog)

    def show_change_path_dialog(self):
        """
        Shows a file browser dialog for changing the root path
        :return: str
        """

        path = self._show_change_path_dialog()
        if path:
            self.set_path(path)
        else:
            self.refresh()

    def show_info_dialog(self, title, text):
        """
        Function that shows an information dialog to the user
        :param title: str
        :param text: str
        :return: QDialogButtonBox.StandardButton
        """

        buttons = QDialogButtonBox.Ok
        return messagebox.MessageBox.question(self, title, text, buttons=buttons)

    def show_question_dialog(self, title, text, buttons=None):
        """
        Function that shows a question dialog to the user
        :param title: str
        :param text: str
        :param buttons: list(QDialogButtonBox.StandardButton)
        :return: QDialogButtonBox.StandardButton
        """

        buttons = buttons or QDialogButtonBox.Yes | QDialogButtonBox.No | QDialogButtonBox.Cancel
        return messagebox.MessageBox.question(self, title, text, buttons=buttons)

    def show_error_dialog(self, title, text):
        """
        Function that shows an error dialog to the user
        :param title: str
        :param text: str
        :return: QDialogButtonBox.StandardButton
        """

        self.show_error_message(text)
        return messagebox.MessageBox.critical(self, title, text)

    def show_exception_dialog(self, title, text):
        """
        Function that shows an exception dialog to the user
        :param title: str
        :param text: str
        :return: QDialogButtonBox.StandardButton
        """

        logger.exception(text)
        self.show_error_dialog(title, text)

    # ============================================================================================================
    # OTHERS
    # ============================================================================================================

    def is_compact_view(self):
        """
        Returns whether the folder and preview widget are hidden
        :return: bool
        """

        return not self.is_sidebar_widget_visible() and not self.is_preview_widget_visible()

    def toggle_view(self):
        """
        Toggles the preview widget and folder widget visibility
        """

        compact = self.is_compact_view()
        if qtutils.is_control_modifier():
            compact = False
            self.set_toolbar_widget_visible(compact)

        self.set_preview_widget_visible(compact)
        self.set_sidebar_widget_visible(compact)

    def update_view_button(self):
        """
        Updates the icon for the view action
        """

        compact = self.is_compact_view()
        action = self.toolbar_widget().find_action('View')
        if not compact:
            new_icon = resources.icon('view_all', theme='black')
        else:
            new_icon = resources.icon('view_compact', theme='black')
        # new_icon.set_color(self.icon_color())
        if new_icon.isNull():
            return
        new_icon.set_color(QColor(255, 255, 255, 255))
        action.setIcon(new_icon)

    # ============================================================================================================
    # SETTINGS
    # ============================================================================================================

    def settings_file_path(self):
        """
        Returns custom settings file path
        :return: str
        """

        return self._settings_file_path

    def set_settings_file_path(self, file_path):
        """
        Sets settings file path used by the library
        :param file_path: str
        """

        self._settings_file_path = file_path

        self._libraries_menu.set_settings_path(self._settings_file_path)

    def settings(self):
        """
        Return a dictionary with the widget settings
        :return: dict
        """

        settings = dict()

        settings['dpi'] = self.dpi()
        settings['kwargs'] = self._kwargs
        settings['library'] = self.library().settings()
        settings['paneSizes'] = self._splitter.sizes()
        settings['trashFolderVisible'] = self.is_trash_folder_visible()
        settings['sidebarWidgetVisible'] = self.is_sidebar_widget_visible()
        settings['previewWidgetVisible'] = self.is_preview_widget_visible()
        settings['toolBarWidgetVisible'] = self.is_toolbar_widget_visible()
        settings['statusBarWidgetVisible'] = self.is_status_widget_visible()
        settings['viewerWidget'] = self.viewer().settings()
        settings['searchWidget'] = self.search_widget().settings()
        settings['sidebarWidget'] = self.sidebar_widget().settings()
        settings['recursiveSearchEnabled'] = self.is_recursive_search_enabled()
        settings['filterByMenu'] = self._filter_by_menu.settings()
        settings['path'] = self.path()

        return settings

    def set_settings(self, settings_dict):
        """
        Sets the widget settings from the given dictionary
        :param settings_dict: dict
        """

        defaults = copy.deepcopy(self.DEFAULT_SETTINGS)
        settings = utils.update(defaults, settings_dict)

        is_refresh_enabled = self.is_refresh_enabled()

        try:
            self.set_refresh_enabled(False)
            self.viewer().set_toast_enabled(False)

            if not self.path():
                path = settings.get('path')
                if path and os.path.exists(path):
                    self.set_path(path)

            dpi = settings.get('dpi', 1.0)
            self.set_dpi(dpi)

            value = settings.get('library')
            if value is not None and self.library():
                self.library().update_settings(value)

            sizes = settings.get('paneSizes')
            if sizes and len(sizes) == 3:
                self.set_sizes(sizes)

            value = settings.get("sidebarWidgetVisible")
            if value is not None:
                self.set_sidebar_widget_visible(value)

            value = settings.get("toolBarWidgetVisible")
            if value is not None:
                self.set_toolbar_widget_visible(value)

            value = settings.get("previewWidgetVisible")
            if value is not None:
                self.set_preview_widget_visible(value)

            value = settings.get("statusBarWidgetVisible")
            if value is not None:
                self.set_status_widget_visible(value)

            value = settings.get('searchWidget')
            if value is not None:
                self.search_widget().set_settings(value)

            value = settings.get("recursiveSearchEnabled")
            if value is not None:
                self.set_recursive_search_enabled(value)

            value = settings.get('filterByMenu')
            if value is not None:
                self._filter_by_menu.set_settings(value)

        finally:
            self.set_refresh_enabled(is_refresh_enabled)
            self.refresh()

        value = settings.get('trashFolderVisible')
        if value is not None:
            self.set_trash_folder_visible(value)

        value = settings.get('sidebarWidget', {})
        self.sidebar_widget().set_settings(value)

        value = settings.get('viewerWidget', {})
        self.viewer().set_settings(value)

        self.update_filters_button()

    @qt_decorators.show_wait_cursor
    def load_settings(self):
        """
        Loads the user settings from disk
        """

        settings = self.read_settings()
        self.set_settings(settings)

    def read_settings(self):
        """
        Reads settings from disk
        :return: dict
        """

        key = self.name()
        data = utils.read_settings(self._settings_file_path)

        return data.get(key, dict())

    def save_settings(self, settings_dict=None):
        """
        Save the settings to the settings path
        :param settings_dict: dict
        """

        settings = utils.read_settings(self._settings_file_path)
        settings.setdefault(self.name(), dict())
        settings[self.name()].update(settings_dict or self.settings())
        utils.save_settings(settings)

        self.show_toast_message('Saved')

    def update_settings(self, settings):
        """
        Saves settings in the settings file on disk
        """

        data = self.read_settings()
        data.update(settings)
        self.save_settings(data)

    def reset_settings(self):
        """
        Reset the settings to the default ones
        """

        self.set_settings(self.DEFAULT_SETTINGS)

    def open_settings(self):
        """
        Opens current library settings file in explorer
        """

        path = self._settings_file_path or utils.settings_path()
        if not path or not os.path.isfile(path):
            return

        fileio.open_browser(path)

    def open_temp_path(self):
        """
        Opens library temporal directory in explorer
        """

        temp_path = utils.temp_path()
        folder.open_folder(temp_path)

    # ============================================================================================================
    # INTERNAL
    # ============================================================================================================

    def _setup_toolbar(self):
        self.add_toolbar_action(
            'New Item', resources.icon('plus'), 'Add a new item to the selected folder',
            callback=self._on_show_new_menu)

        self._toolbar_widget.addWidget(self._search_widget)

        self.add_toolbar_action(
            'Filters', resources.icon('filter'),
            'Filter the current results by type.\nCtrl + Click will hide the others and show the selected one.',
            callback=self._on_show_filter_by_menu)

        self.add_toolbar_action(
            'Item View', resources.icon('slider'), 'Change the style of the item view',
            callback=self._on_show_item_view_menu)

        self.add_toolbar_action(
            'Group By', resources.icon('grid_view'), 'Group the current items in the view by column',
            callback=self._on_show_group_by_menu)

        self.add_toolbar_action(
            'Sort By', resources.icon('descending_sorting'), 'Group the current items in the view by column',
            callback=self._on_show_sort_by_menu)

        self.add_toolbar_action(
            'View', resources.icon('add'),
            'Choose to show/hide both the preview and navigation pane\nCtrl + click will hide the menu bar as well.',
            callback=self._on_toggle_view)

        self.add_toolbar_action(
            'Sync Items', resources.icon('sync'), 'Sync with the filesystem', callback=self._on_sync)

        self.add_toolbar_action(
            'Settings', resources.icon('settings'), 'Settings menu', callback=self._on_show_settings_menu)

    def _create_new_item_menu(self):
        """
        Internal function that creates a new item menu for adding new items
        :return: QMenu
        """

        item_icon = resources.icon('add')
        menu = QMenu(self)
        menu.setIcon(item_icon)
        menu.setTitle('New')

        library = self.library()
        if not library:
            return

        all_sorted_data = sorted(list(library.get_all_data_plugins()), key=operator.attrgetter('PRIORITY'))

        for data_item_class in all_sorted_data:
            action = self.create_action(data_item_class, menu, self)
            if action:
                action_icon = icon.Icon(action.icon())
                # action_icon.set_color(self.icon_color())
                action.setIcon(action_icon)
                menu.addAction(action)

        return menu

    def _set_progress_bar_value(self, label, value=-1):
        """
        Internal function that sets the progress bar label and value
        :param label: str
        :param value: int
        """

        progress_bar = self.status_widget().progress_bar()
        if value == -1:
            value = 100

        progress_bar.set_value(value)
        progress_bar.set_text(label)

    def _create_item_context_menu(self, items):
        """
        Internal function that returns the item context menu for the given items
        :param items: list(LibraryItem)
        :return:
        """

        self._menu_items = list()

        context_menu = QMenu(self)

        item_view = None
        if items:
            item_data = items[-1]
            if isinstance(item_data, items_view.ItemView):
                item_view = item_data
            else:
                view_class = self._items_factory.get_view(item_data)
                if view_class:
                    item_view = view_class(item_data, library_window=self)
            if item_view:
                item_view.context_menu(context_menu)

                # NOTE: We do thos to avoid Python to garbage collect the item views. Otherwise menu functionality
                # NOTE: related with item views will not work
                self._menu_items.append(item_view)

        if not self.is_locked():
            context_menu.addMenu(self._create_new_item_menu())
            if item_view:
                edit_icon = resources.icon('edit')
                edit_menu = QMenu(self)
                edit_menu.setTitle('Edit')
                edit_menu.setIcon(edit_icon)
                context_menu.addMenu(edit_menu)
                item_view.context_edit_menu(edit_menu)
                if self.trash_enabled():
                    edit_menu.addSeparator()
                    callback = partial(self.show_move_items_to_trash_dialog, items)
                    action = QAction(resources.icon('trash'), 'Move to Trash', edit_menu)
                    action.setEnabled(not self.is_trash_selected())
                    action.triggered.connect(callback)
                    edit_menu.addAction(action)

        context_menu.addSeparator()
        context_menu.addMenu(self._create_settings_menu())

        return context_menu

    def _create_settings_menu(self):
        """
        Returns the settings menu for changing the library widget
        :return: QMenu
        """

        settings_icon = resources.icon('settings')
        context_menu = menu.Menu('', self)
        context_menu.setTitle('Settings')
        context_menu.setIcon(settings_icon)

        settings_action = context_menu.addAction(settings_icon, 'Settings')
        settings_action.triggered.connect(self._on_show_settings_dialog)

        sync_action = context_menu.addAction(resources.icon('sync'), 'Sync')
        sync_action.triggered.connect(self._on_sync)

        context_menu.addSeparator()

        self._libraries_menu = libraries.LibrariesMenu(self._settings_file_path, library_window=self)
        context_menu.addMenu(self._libraries_menu)

        context_menu.addSeparator()

        # if consts.DPI_ENABLED:
        #     dpi_action = action.SliderAction('Dpi', context_menu)
        #     dpi = self.dpi() * 100.0
        #     dpi_action.slider().setRange(consts.DPI_MIN_VALUE, consts.DPI_MAX_VALUE)
        #     dpi_action.slider().setValue(dpi)
        #     dpi_action.valueChanged.connect(self._on_dpi_slider_changed)
        #     context_menu.addAction(dpi_action)
        #     context_menu.addSeparator()

        show_menu_action = QAction('Show Menu', context_menu)
        show_menu_action.setCheckable(True)
        show_menu_action.setChecked(self.is_toolbar_widget_visible())
        show_menu_action.triggered[bool].connect(self.set_toolbar_widget_visible)
        context_menu.addAction(show_menu_action)

        show_sidebar_action = QAction('Show Sidebar', context_menu)
        show_sidebar_action.setCheckable(True)
        show_sidebar_action.setChecked(self.is_sidebar_widget_visible())
        show_sidebar_action.triggered[bool].connect(self.set_sidebar_widget_visible)
        context_menu.addAction(show_sidebar_action)

        show_preview_action = QAction('Show Preview', context_menu)
        show_preview_action.setCheckable(True)
        show_preview_action.setChecked(self.is_preview_widget_visible())
        show_preview_action.triggered[bool].connect(self.set_preview_widget_visible)
        context_menu.addAction(show_preview_action)

        show_status_action = QAction('Show Status', context_menu)
        show_status_action.setCheckable(True)
        show_status_action.setChecked(self.is_status_widget_visible())
        show_status_action.triggered[bool].connect(self.set_status_widget_visible)
        context_menu.addAction(show_status_action)

        context_menu.addSeparator()

        save_settings_action = QAction(resources.icon('save'), 'Save Settings', context_menu)
        save_settings_action.triggered.connect(self.save_settings)
        context_menu.addAction(save_settings_action)

        reset_settings_action = QAction(resources.icon('reset'), 'Reset Settings', context_menu)
        reset_settings_action.triggered.connect(self.reset_settings)
        context_menu.addAction(reset_settings_action)

        open_settings_action = QAction(resources.icon('open'), 'Open Settings', context_menu)
        open_settings_action.triggered.connect(self.open_settings)
        context_menu.addAction(open_settings_action)

        if self.TEMP_PATH_MENU_ENABLED:
            temp_path_action = QAction(resources.icon('open'), 'Open Temp Path', context_menu)
            temp_path_action.triggered.connect(self.open_temp_path)
            context_menu.addAction(temp_path_action)

        context_menu.addSeparator()

        if self.trash_enabled():
            show_trash_action = QAction(resources.icon('trash'), 'Show Trash Folder', context_menu)
            show_trash_action.setEnabled(self.trash_folder_exists())
            show_trash_action.setCheckable(True)
            show_trash_action.setChecked(self.is_trash_folder_visible())
            show_trash_action.triggered[bool].connect(self.set_trash_folder_visible)
            context_menu.addAction(show_trash_action)

        context_menu.addSeparator()

        recursive_search_action = QAction(resources.icon('repeat'), 'Enable Recursive Search', context_menu)
        recursive_search_action.setCheckable(True)
        recursive_search_action.setChecked(self.is_recursive_search_enabled())
        recursive_search_action.triggered[bool].connect(self.set_recursive_search_enabled)
        context_menu.addAction(recursive_search_action)

        context_menu.addSeparator()

        cleanup_library_action = QAction(resources.icon('clean'), 'Clean Library', context_menu)
        cleanup_library_action.triggered[bool].connect(self.clean_library)
        context_menu.addAction(cleanup_library_action)

        context_menu.addSeparator()

        view_menu = self.viewer().create_settings_menu()
        context_menu.addMenu(view_menu)

        return context_menu

    @qt_decorators.show_arrow_cursor
    def _show_change_path_dialog(self):
        """
        Internal function that opens a file dialog for setting a new root path
        :return: str
        """

        path = self.path()
        if not path:
            path = os.path.expanduser('~')

        path = QFileDialog.getExistingDirectory(None, 'Choose a root folder location', path)
        if not path:
            return

        data_path = path_utils.join_path(path, 'data.db')

        return path_utils.clean_path(data_path)

    # ============================================================================================================
    # CALLBACKS
    # ============================================================================================================

    def _on_show_new_menu(self):
        """
        Internal callback function that is called when user right clicks on an item
        Shows items contextual menu
        :return: QMenu
        """

        new_menu = self._create_new_item_menu()
        if not new_menu:
            return
        point = self.toolbar_widget().mapToGlobal(self.toolbar_widget().rect().bottomLeft())
        new_menu.show()

        return new_menu.exec_(point)

    def _on_show_filter_by_menu(self):
        """
        Internal callback function called when the user clicks on the filter action
        """

        widget = self.toolbar_widget().find_tool_button('Filters')
        point = widget.mapToGlobal(QPoint(0, widget.height()))
        self._filter_by_menu.show(point)
        self.update_filters_button()

    def _on_show_item_view_menu(self):
        """
        Internal callback function when triggered when the user clicks on item view action
        """

        menu = self.viewer().create_settings_menu()
        item_view_widget = self.toolbar_widget().find_tool_button('Item View')
        point = item_view_widget.mapToGlobal(QPoint(0, item_view_widget.height()))
        menu.exec_(point)

    def _on_show_group_by_menu(self):
        """
        Internal callback function called when the user presses group by action
        """

        widget = self.toolbar_widget().find_tool_button('Group By')
        point = widget.mapToGlobal(QPoint(0, widget.height()))
        self._group_by_menu.show(point)

    def _on_show_sort_by_menu(self):
        """
        Internal callback function called when the user presses Sort by action
        """

        widget = self.toolbar_widget().find_tool_button('Sort By')
        point = widget.mapToGlobal(QPoint(0, widget.height()))
        self._sort_by_menu.show(point)

    def _on_toggle_view(self):
        """
        Internal callback function called whe the user press view action
        """

        self.toggle_view()

    def _on_show_settings_menu(self):
        """
        Internal callback function triggered when the user show settings
        Show the settings menu at the current cursor position
        :return: QAction
        """

        settings_menu = self._create_settings_menu()
        point = self.toolbar_widget().rect().bottomRight()
        point = self.toolbar_widget().mapToGlobal(point)
        if osplatform.is_linux():
            settings_menu.move(-1000, -1000)
        settings_menu.show()
        point.setX(point.x() - settings_menu.width())

        return settings_menu.exec_(point)

    def _on_sync(self):
        """
        Internal callback function that is executed when the user selects the Sync
        context menu action
        """

        self.sync(force_search=True)

    def _on_show_settings_dialog(self):
        # TODO: Open settings dialog
        # If there is an attacher, we use attacher settings view, if not we create it
        pass

    def _on_folder_selection_changed(self):
        """
        Internal callback triggered when a folder is selected or deselect from sidebar widget
        """

        item_path = self.selected_folder_path()
        # self.library().search()
        self.folderSelectionChanged.emit(item_path)
        # self.globalSignal.folderSelectionChanged.emit(self, item_path)

    def _on_item_dropped(self, event):
        """
        Internal callback function that is triggered when items are dropped on the viewer or sidebar widget
        :param event: QEvent
        """

        mime_data = event.mimeData()
        if mime_data.hasUrls():
            urls = mime_data.urls()
            path = self.selected_folder_path()
            items = self.create_items_from_urls(urls)
            if self.is_move_items_enabled():
                self.show_move_items_dialog(items, target_folder=path)
            elif not self.is_custom_order_enabled():
                self.show_info_message('Please use sort by "Custom Order" to reorder items!')

    def _on_show_folders_menu(self):
        """
        Internal callback function that is called when sidebar context menu is requested
        :return: QAction
        """

        folders_menu = self.create_folder_context_menu()
        point = QCursor.pos()
        point.setX(point.x() + 3)
        point.setY(point.y() + 3)
        action = folders_menu.exec_(point)
        folders_menu.close()

        return action

    def _test(self):
        print('hello')

    def _on_sidebar_settings_menu_requested(self, settings_menu):
        """
        Internal callback function triggered when sidebar settings menu is requested to show
        :param settings_menu: Menu
        """

        change_path_action = QAction(resources.icon('change'), 'Change Path', settings_menu)
        change_path_action.triggered.connect(self.show_change_path_dialog)
        settings_menu.addAction(change_path_action)
        settings_menu.addSeparator()

    def _on_item_selection_changed(self):
        """
        Internal callback function that is triggered when an item is selected or deselected
        """

        item = self.viewer().selected_item()

        self.set_preview_widget_from_item(item)
        self.itemSelectionChanged.emit(item)

    def _on_key_pressed(self, event):
        """
        Internal callback function that is triggered when a key is pressed on viewer widget
        :param event: QKeyEvent
        """

        text = event.text().strip()

        if not text.isalpha() and not text.isdigit():
            return

        if text and not self.search_widget().hasFocus():
            self.search_widget().setFocus()
            self.search_widget().setText(text)

    def _on_item_moved(self, item):
        """
        Internal callback function that is triggered when the custom order has changed
        :param item, LibraryItem
        """

        self.save_custom_order()

    def _on_item_dropped(self, event):
        """
        Internal callback function that is triggered when items are dropped on the viewer or sidebar widget
        :param event: QEvent
        """

        mime_data = event.mimeData()
        if mime_data.hasUrls():
            urls = mime_data.urls()
            path = self.selected_folder_path()
            items = self.create_items_from_urls(urls)
            if self.is_move_items_enabled():
                self.show_move_items_dialog(items, target_folder=path)
            elif not self.is_custom_order_enabled():
                self.show_info_message('Please use sort by "Custom Order" to reorder items!')

    def _on_show_items_context_menu(self):
        """
        Internal callback function that is called when user right clicks on muscle viewer
        Shows the item context menu at the current cursor position
        :param pos, QPoint
        :return: QAction
        """

        item_views = self.viewer().selected_items()
        menu = self._create_item_context_menu(item_views)
        point = QCursor.pos()
        point.setX(point.x() + 3)
        point.setY(point.y() + 3)
        action = menu.exec_(point)

        return action



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains consts definitions used by data library
"""

from __future__ import print_function, division, absolute_import

from Qt.QtGui import QColor

TOOL_ID = 'tpDcc-tools-datalibrary'

DEFAULT_LIBRARY_NAME = 'Default'
DEFAULT_RECURSIVE_SEARCH_ENABLED = False
DEFAULT_RECURSIVE_DEPTH = 8

ITEM_DEFAULT_THUMBNAIL_NAME = 'thumbnail.jpg'
ITEM_DEFAULT_MAX_ICON_SIZE = 256
ITEM_DEFAULT_FONT_SIZE = 12
ITEM_DEFAULT_PLAYHEAD_COLOR = QColor(255, 255, 255, 220)
ITEM_DEFAULT_THUMBNAIL_COLUMN = 0
ITEM_DEFAULT_ENABLE_THUMBNAIL_THREAD = True

GROUP_ITEM_DEFAULT_FONT_SIZE = 24
GROUP_ITEM_PADDING_LEFT = 2
GROUP_ITEM_PADDING_RIGHT = 20
GROUP_ITEM_HEIGHT = 28

PROGRESS_BAR_VISIBLE = True

ITEM_DEFAULT_MENU_ICON = ''
ITEM_DEFAULT_MENU_ORDER = 10
ITEM_DEFAULT_EXTENSION = ''

TRASH_NAME = 'trash'
TRASH_ENABLED = True
ICON_BADGE_COLOR = QColor(230, 230, 0)

DEFAULT_ICON_MODE = 'icon'
DEFAULT_TABLE_MODE = 'table'
VIEWER_DEFAULT_PADDING = 5
VIEWER_DEFAULT_ZOOM_AMOUNT = 90
VIEWER_DEFAULT_TEXT_HEIGHT = 20
VIEWER_DEFAULT_WHEEL_SCROLL_STEP = 2
VIEWER_DEFAULT_MIN_SPACING = 0
VIEWER_DEFAULT_MAX_SPACING = 50
VIEWER_DEFAULT_MIN_LIST_SIZE = 15
VIEWER_DEFAULT_MIN_ICON_SIZE = 50
VIEWER_DEFAULT_TEXT_COLOR = QColor(255, 255, 255, 200)
VIEWER_DEFAULT_SELECTED_TEXT_COLOR = QColor(255, 255, 255, 200)
VIEWER_DEFAULT_BACKGROUND_COLOR = QColor(255, 255, 255, 30)
VIEWER_DEFAULT_BACKGROUND_HOVER_COLOR = QColor(255, 255, 255, 35)
VIEWER_DEFAULT_BACKGROUND_SELECTED_COLOR = QColor(30, 150, 255)

LIST_DEFAULT_DRAG_THRESHOLD = 10

TREE_MINIMUM_WIDTH = 5
TREE_DEFAULT_WIDTH = 100

DEFAULT_FOLDER_ICON_COLOR = 'rgb(150, 150, 150, 100)'
DEFAULT_FOLDER_ICON_COLORS = [
    "rgb(239, 112, 99)",
    "rgb(239, 207, 103)",
    "rgb(136, 200, 101)",
    "rgb(111, 183, 239)",
    "rgb(199, 142, 220)",
    DEFAULT_FOLDER_ICON_COLOR
]

DEFAULT_FOLDER_ICONS = [
    'folder', 'user', 'character', 'inbox', 'favorite_folder', 'movie', 'camera', 'asset', 'assets', 'cloud',
    'book', 'archive_folder', 'archive', 'circle', 'share', 'tree', 'mountain', 'trash', 'layers', 'database',
    'video', 'face', 'hand', 'globe'
]



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains base library data item widget view implementation
"""

from __future__ import print_function, division, absolute_import

import os
import math

from Qt.QtCore import Qt, Signal, QObject, QRect, QSize, QThreadPool, QUrl
from Qt.QtWidgets import QApplication, QStyle, QTreeWidget, QTreeWidgetItem
from Qt.QtGui import QFontMetrics, QColor, QIcon, QPixmap, QPen, QBrush, QMovie

from tpDcc import dcc
from tpDcc.managers import resources
from tpDcc.libs.python import python, path as path_utils
from tpDcc.libs.resources.core import theme
from tpDcc.libs.qt.core import qtutils, image

from tpDcc.tools.datalibrary.core import consts


class LabelDisplayOption:

    Hide = "hide label"
    Over = "label over item"
    Under = "label under item"

    @staticmethod
    def values():
        return [
            LabelDisplayOption.Hide,
            LabelDisplayOption.Over,
            LabelDisplayOption.Under
        ]


class GlobalDataItemSignals(QObject, object):
    blendChanged = Signal(float)
    loadValueChanged = Signal(object, object)


@theme.mixin
class ItemView(QTreeWidgetItem):

    NAME = 'Item View'
    VERSION = 0
    PRIORITY = 0
    REPRESENTING = []           # List of data types this view can represent

    THREAD_POOL = QThreadPool()

    MAX_ICON_SIZE = consts.ITEM_DEFAULT_MAX_ICON_SIZE
    DEFAULT_FONT_SIZE = consts.ITEM_DEFAULT_FONT_SIZE
    DEFAULT_PLAYHEAD_COLOR = consts.ITEM_DEFAULT_PLAYHEAD_COLOR

    DEFAULT_THUMBNAIL_NAME = 'tpDcc.png'
    DEFAULT_THUMBNAIL_COLUMN = consts.ITEM_DEFAULT_THUMBNAIL_COLUMN
    ENABLE_THUMBNAIL_THREAD = consts.ITEM_DEFAULT_ENABLE_THUMBNAIL_THREAD

    PAINT_SLIDER = False
    _TYPE_PIXMAP_CACHE = dict()

    _globalSignals = GlobalDataItemSignals()
    blendChanged = _globalSignals.blendChanged
    loadValueChanged = _globalSignals.loadValueChanged

    def __init__(self, data_item, *args, **kwargs):
        super(ItemView, self).__init__(*args, **kwargs)

        self._size = None
        self._rect = None
        self._text_column_order = list()

        self._item = data_item

        self._icon = dict()
        self._thumbnail_icon = None
        self._fonts = dict()
        self._thread = None
        self._pixmap = dict()
        self._pixmap_rect = None
        self._pixmap_scaled = None
        self._type_pixmap = None

        self._mime_text = None
        self._drag_enabled = True
        self._under_mouse = False
        self._search_text = None
        self._info_widget = None
        self._viewer = None
        self._stretch_to_widget = None

        self._group_item = None
        self._group_column = 0

        self._image_sequence = None
        self._image_sequence_path = ''

        self._blend_down = False
        self._blend_value = 0.0
        self._blend_prev_value = 0.0
        self._blend_position = None
        self._blending_enabled = False

        self._worker = image.ImageWorker()
        self._worker.setAutoDelete(False)
        self._worker.signals.triggered.connect(self._on_thumbnail_from_image)
        self._worker_started = False

        self._icon_path = None

        icons_path = path_utils.join_path('icons', self.theme().name().lower())
        color_icons_path = path_utils.join_path('icons', 'color')

        icon_name = data_item.icon() if data_item else None
        if not icon_name:
            icon_name = self.DEFAULT_THUMBNAIL_NAME
        else:
            icon_name, icon_extension = os.path.splitext(icon_name)
            if not icon_extension:
                icon_name = '{}.png'.format(icon_name)

        dcc_name = dcc.client().get_name()
        type_icon = icon_name if icon_name == dcc_name + '.png' else None
        self._type_icon_path = resources.get(color_icons_path, type_icon) if type_icon else ''

        self._default_thumbnail_path = resources.get(
            icons_path, icon_name) or resources.get(icons_path, self.DEFAULT_THUMBNAIL_NAME)

    def __eq__(self, other):
        return id(other) == id(self)

    # Necessary for Python compatibility
    # https://stackoverflow.com/questions/1608842/types-that-define-eq-are-unhashable
    def __hash__(self):
        return id(self)

    def __ne__(self, other):
        return id(other) != id(self)

    def __del__(self):
        """
        When the object is deleted we make sure the sequence is stopped
        """

        self.stop()

    # =================================================================================================================
    # PROPERTIES
    # =================================================================================================================

    @property
    def item(self):
        return self._item

    # =================================================================================================================
    # OVERRIDES
    # =================================================================================================================

    def sizeHint(self, column=0):
        """
        Returns the current size of the item
        :param column: int
        :return: QSize
        """

        if self.stretch_to_widget():
            if self._size:
                size = self._size
            else:
                size = self.viewer().icon_size()
            w = self.stretch_to_widget().width()
            h = size.height()
            return QSize(w - 20, h)

        if self._size:
            return self._size
        else:
            icon_size = self.viewer().icon_size()
            if self.is_label_under_item():
                w = icon_size.width()
                h = icon_size.width() + self.text_height()
                icon_size = QSize(w, h)

            return icon_size

    def setHidden(self, value):
        """
        Overrides base QTreeWidgetItem.setHidden function
        Set the item hidden
        :param value: bool
        """

        super(ItemView, self).setHidden(value)
        row = self.treeWidget().index_from_item(self).row()
        self.viewer().list_view().setRowHidden(row, value)

    def backgroundColor(self):
        """
        Returns the background color for the item
        :return: QColor
        """

        return self.viewer().background_color()

    def icon(self, column):
        """
        Overrides base QTreeWidgetItem icon function
        Overrides icon to add support for thumbnail icon
        :param column: int
        :return: QIcon
        """

        item_icon = QTreeWidgetItem.icon(self, column)
        if not item_icon and column == self.DEFAULT_THUMBNAIL_COLUMN:
            item_icon = self.thumbnail_icon()

        return item_icon

    def setIcon(self, column, icon, color=None):
        """
        Overrides base QTreeWidgetItem setIcon function
        :param column: int or str
        :param icon: QIcon
        :param color: QColor or None
        """

        is_app_running = bool(QApplication.instance())
        if not is_app_running:
            return

        if python.is_string(icon):
            if not os.path.exists(icon):
                color = color or QColor(255, 255, 255, 20)
                icon = resources.icon('image', color=color)
            else:
                icon = QIcon(icon)
        if python.is_string(column):
            self._icon[column] = icon
        else:
            self._pixmap[column] = None
            super(ItemView, self).setIcon(column, icon)

        self.update_icon()

    def setFont(self, column, font):
        """
        Overrides base QTreeWidgetItem setFont function
        Sets the font for the given column
        :param column: int
        :param font: QFont
        """

        self._fonts[column] = font

    def textAlignment(self, column):
        """
        Returns the text alinment for the label in the given column
        :param column: int
        :return: QAlignmentFlag
        """

        if self.viewer().is_icon_view():
            return Qt.AlignCenter
        else:
            return QTreeWidgetItem.textAlignment(self, column)

    # =================================================================================================================
    # BASE
    # =================================================================================================================

    def name(self):
        """
        Returns item data name
        :return: str
        """

        return str(self.item.data.get('name'))

    def display_text(self, label):
        """
        Returns the sort data for the given column
        :param label: str
        :return: str
        """

        return str(self.item.data.get(label, ''))

    def sort_text(self, label):
        """
        Returns the sort data for the given column
        :param label: str
        :return: str
        """

        return str(self.item.data.get(label, ''))

    def stretch_to_widget(self):
        """
        Returns the stretch to widget widget
        :return: QWidget
        """

        return self._stretch_to_widget

    def set_stretch_to_widget(self, widget):
        """
        Sets the width of the item to the width of the given widget
        :param widget: QWidget
        """

        self._stretch_to_widget = widget

    def set_size(self, size):
        """
        Sets the size for the item
        :param size: QSize
        """

        self._size = size

    def pixmap(self, column):
        """
        Returns the pixmap for the given column
        :param column: int
        :return: QPixmap
        """

        if not self._pixmap.get(column):
            icon = self.icon(column)
            if icon:
                size = QSize(self.MAX_ICON_SIZE, self.MAX_ICON_SIZE)
                icon_size = icon.actualSize(size)
                self._pixmap[column] = icon.pixmap(icon_size)

        return self._pixmap.get(column)

    def set_pixmap(self, column, pixmap):
        """
        Sets the pixmap to be displayed in the given column
        :param column: int
        :param pixmap: QPixmap
        """

        self._pixmap[column] = pixmap

    def icon_path(self):
        """
        Returns icon path for current item
        :return: str
        """

        return self._icon_path

    def set_icon_path(self, path):
        """
        Sets icon path for the item
        :param path: str
        """

        self._icon_path = path

    def take_from_tree(self):
        """
        Takes this item from the tree
        """

        tree = self.treeWidget()
        parent = self.parent()
        if parent:
            parent.takeChild(parent.indexOfChild(self))
        else:
            tree.takeTopLevelItem(tree.indexOfTopLevelItem(self))

    def group_item(self):
        """
        Returns current group item
        :return: LibraryGroupItem
        """

        return self._group_item

    def set_group_item(self, group_item):
        """
        Sets current group item
        :param group_item: LibraryGroupItem
        """

        self._group_item = group_item

    def update_icon(self):
        """
        Clears the pixmap cache for the item
        """

        self.clear_cache()

    def clear_cache(self):
        """
        Clears the thumbnail cache
        """

        self._pixmap = dict()
        self._pixmap_rect = None
        self._pixmap_scaled = None
        self._thumbnail_icon = None

    def update(self):
        """
        Refresh teh visual state of the icon
        """

        self.update_icon()
        self.update_frame()

    def context_menu(self, menu, items=None):
        """
        Returns the context menu for the item
        This function must be implemented in a subclass to return a custom context menu for the item
        :param menu: QMenu
        :param items: list
        """

        pass

    # ============================================================================================================
    # MOUSE / KEYBOARD
    # ============================================================================================================

    def under_mouse(self):
        """
        Returns whether the items is under the mouse cursor or not
        :return: bool
        """

        return self._under_mouse

    def drop_event(self, event):
        """
        Reimplement in subclass to receive drop items for the item
        :param event: QDropEvent
        """

        pass

    def mouse_enter_event(self, event):
        """
        Reimplement in subclass to receive mouse enter events for the item
        :param event: QMouseEvent
        """

        self._under_mouse = True
        self.play()

    def mouse_leave_event(self, event):
        """
        Reimplement in subclass to receive mouse leave events for the item
        :param event: QMouseEvent
        """

        self._under_mouse = False
        self.stop()

    def mouse_move_event(self, event):
        """
        Reimplement in subclass to receive mouse move events for the item
        :param event: QMouseEvent
        """

        self.blending_event(event)
        self.image_sequence_event(event)

    def mouse_press_event(self, event):
        """
        Reimplement in subclass to receive mouse press events for the item
        :param event: QMouseEvent
        """

        if event.button() == Qt.MidButton:
            if self.is_blending_enabled():
                self.set_is_blending(True)
                self._blend_position = event.pos()

    def mouse_release_event(self, event):
        """
        Reimplement in subclass to receive mouse release events for the item
        :param event: QMouseEvent
        """

        if self.is_blending():
            self._blend_position = None
            self._blend_prev_value = self.blend_value()

    def key_press_event(self, event):
        """
        Reimplement in subclass to receive key press events for the item
        :param event: QKeyEvent
        """

        pass

    def key_release_event(self, event):
        """
        Reimplement in subclass to receive key release events for the item
        :param event: QKeyEvent
        """

        pass

    def clicked(self):
        """
        Triggered when an item is clicked
        """

        pass

    def double_clicked(self):
        """
        Triggered when an item is double clicked
        """

        pass

    def selection_changed(self):
        """
        Triggered when an item has been either selected or deselected
        """

        self.reset_blending()

    # ============================================================================================================
    # THUMBNAIL
    # ============================================================================================================

    def default_thumbnail_path(self):
        """
        Returns the default thumbnail path
        :return: str
        """
        return self._default_thumbnail_path

    def default_thumbnail_icon(self):
        """
        Returns the default thumbnail icon
        :return: QIcon
        """

        return QIcon(self.default_thumbnail_path())

    def type_icon_path(self):
        """
        Returns the type icon path on disk
        :return: str
        """

        if not self._type_icon_path:
            return self._icon_path

        return self._type_icon_path

    def is_default_thumbnail_path(self):
        """
        Returns whether or not current thumbnail path is the default one
        :return: bool
        """

        return self.thumbnail_path() == self._default_thumbnail_path

    def thumbnail_path(self):
        """
        Return the thumbnail path for the item on disk
        :return: str
        """

        item_path = self.item.format_identifier() if self.item else None
        if not item_path:
            return self._default_thumbnail_path

        thumbnail_path = os.path.dirname(item_path) if os.path.isfile(item_path) else item_path
        thumbnail_path = path_utils.join_path(thumbnail_path, consts.ITEM_DEFAULT_THUMBNAIL_NAME)
        if os.path.isfile(thumbnail_path):
            return thumbnail_path

        thumbnail_path = thumbnail_path.replace('.jpg', '.png')
        if os.path.isfile(thumbnail_path):
            return thumbnail_path

        return self._default_thumbnail_path

    def thumbnail_icon(self):
        """
        Returns the thumbnail icon
        :return: QIcon
        """

        thumbnail_path = self.thumbnail_path()
        if not self._thumbnail_icon:
            if self.ENABLE_THUMBNAIL_THREAD and not self._worker_started:
                self._worker_started = True
                self._worker.set_path(thumbnail_path)
                self.THREAD_POOL.start(self._worker)
                self._thumbnail_icon = self.default_thumbnail_icon()
            else:
                self._thumbnail_icon = QIcon(thumbnail_path)

        return self._thumbnail_icon

    # =================================================================================================================
    # SEQUENCE
    # =================================================================================================================

    def image_sequence(self):
        """
        Return ImageSequence of the item
        :return: image.ImageSequence or QMovie
        """

        return self._image_sequence

    def set_image_sequence(self, image_sequence):
        """
        Set the image sequence of the item
        :param image_sequence: image.ImageSequence or QMovie
        """

        self._image_sequence = image_sequence

    def image_sequence_path(self):
        """
        Return the path where image sequence is located on disk
        :return: str
        """

        return self._image_sequence_path

    def set_image_sequence_path(self, path):
        """
        Set the path where image sequence is located on disk
        :param path: str
        """

        self._image_sequence_path = path

    def reset_image_sequence(self):
        """
        Reset image sequence
        """

        self._image_sequence = None

    def play(self):
        """
        Start play image sequence
        """

        self.reset_image_sequence()
        path = self.image_sequence_path() or self.thumbnail_path()
        movie = None

        if not path:
            return

        if os.path.isfile(path) and path.lower().endswith('.gif'):
            movie = QMovie(path)
            movie.setCacheMode(QMovie.CacheAll)
            movie.frameChanged.connect(self._on_frame_changed)
        elif os.path.isdir(path):
            if not self.image_sequence():
                movie = image.ImageSequence(path)
                movie.frameChanged.connect(self._on_frame_changed)

        if movie:
            self.set_image_sequence(movie)
            self.image_sequence().start()

    def update_frame(self):
        """
        Function that updates the current frame
        """

        if self.image_sequence():
            pixmap = self.image_sequence().current_pixmap()
            self.setIcon(0, pixmap)

    def stop(self):
        """
        Stop play image sequence
        """

        if self._image_sequence:
            self._image_sequence.stop()

    def playhead_color(self):
        """
        Returns playehad color
        :return: QColor
        """

        return self.DEFAULT_PLAYHEAD_COLOR

    def image_sequence_event(self, event):
        """
        :param event: QEvent
        """

        if not self.image_sequence() or not qtutils.is_control_modifier() or not self.rect():
            return

        x = event.pos().x() - self.rect().x()
        width = self.rect().width()
        percent = 1.0 - (float(width - x) / float(width))
        frame = int(self.image_sequence().frameCount() * percent)
        self.image_sequence().jumpToFrame(frame)
        self.update_frame()

    # =================================================================================================================
    # DRAG & DROP
    # =================================================================================================================

    def drag_enabled(self):
        """
        Return whether the item can be dragged or not
        :return: bool
        """

        return self._drag_enabled

    def set_drag_enabled(self, flag):
        """
        Set whether item can be dragged or not
        :param flag: bool
        """

        self._drag_enabled = flag

    def mime_text(self):
        """
        Returns the mime text for drag and drop
        :return: str
        """

        return self._mime_text or self.text(0)

    def set_mime_text(self, text):
        """
        Sets the mime text for drag and drop
        :param text: str
        """

        self._mime_text = text

    def url(self):
        """
        Used by the mime data when dragging/droping the item
        :return: Qurl
        """

        if not self._url:
            self._url = QUrl(self.text(0))

        return self._url

    def set_url(self, url):
        """
        Sets the url object of the current item
        :param url: QUrl or None
        """

        self._url = url

    # =================================================================================================================
    # SEARCH
    # =================================================================================================================

    def search_text(self):
        """
        Returns the search string used for finding the item
        :return: str
        """

        if not self._search_text:
            self._search_text = str(self._data)

        return self._search_text

    # =================================================================================================================
    # VIEWER WIDGET
    # =================================================================================================================

    def viewer(self):
        """
        Returns the viewer widget that contains the item
        :return: libraryViewer
        """

        viewer_widget = None
        if self.treeWidget():
            viewer_widget = self.treeWidget().parent()

        return viewer_widget

    def dpi(self):
        """
        Return current dpi
        :return: int
        """

        if self.viewer():
            return self.viewer().dpi()

        return 1

    def padding(self):
        """
        Returns the padding/border size for the item
        :return: int
        """

        return self.viewer().padding()

    def text_height(self):
        """
        Returns the height of the text for the item
        :return: int
        """

        return self.viewer().item_text_height()

    def label_display_option(self):
        """
        Returns label display option of the item
        :return: LabelDisplayOption
        """

        return self.viewer().label_display_option()

    def is_text_visible(self):
        """
        Returns whether the text is visible or not
        :return: bool
        """

        return self.label_display_option() != LabelDisplayOption.Hide

    def is_label_over_item(self):
        """
        Returns whether or not item label should be displayed over the item
        :return: bool
        """

        return self.label_display_option() == LabelDisplayOption.Over

    def is_label_under_item(self):
        """
        Returns whether or not item label should be displayed under the item
        :return: bool
        """

        return self.label_display_option() == LabelDisplayOption.Under

    # =================================================================================================================
    # TREE WIDGET
    # =================================================================================================================

    def column_from_label(self, label):
        """
        Returns column fro mlabel
        :param label: str
        :return: str
        """

        if self.treeWidget():
            return self.treeWidget().column_from_label(label)

        return None

    def label_from_column(self, column):
        """
        Returns label of the given column
        :param column: int
        :return: str
        """

        if self.treeWidget():
            return self.treeWidget().label_from_column(column)

        return None

    # ============================================================================================================
    # PAINT
    # ============================================================================================================

    def type_pixmap(self):
        """
        Returns the type pixmap for the item data type
        :return: QPixmap
        """

        path = self.type_icon_path()
        pixmap = self._TYPE_PIXMAP_CACHE.get(path)
        if not pixmap and path and os.path.isfile(path):
            self._TYPE_PIXMAP_CACHE[path] = QPixmap(path)

        return self._TYPE_PIXMAP_CACHE.get(path)

    def font_size(self):
        """
        Returns the font size for the item
        :return: int
        """

        return self.DEFAULT_FONT_SIZE

    def font(self, column):
        """
        Returns the font for the given column
        :param column: int
        :return: QFont
        """

        default = QTreeWidgetItem.font(self, column)
        font = self._fonts.get(column, default)
        font.setPixelSize(self.font_size() * self.dpi())

        return font

    def set_font(self, column, font):
        """
        Sets the font used by the items in the given column
        :param column: int
        :param font: QFont
        """

        self._fonts[column] = font

    def text_width(self, column):
        """
        Returns the text width of the given column
        :param column: int
        :return: int
        """

        text = self.text(column)
        font = self.font(column)
        metrics = QFontMetrics(font)
        text_width = metrics.width(text)

        return text_width

    def text_color(self):
        """
        Returns the text color for the item
        :return: QColor
        """

        # TODO: change to use foreground role
        # return self.itemsWidget().palette().color(self.itemsWidget().foregroundRole())
        return self.viewer().text_color()

    def text_selected_color(self):
        """
        Returns the selected txt color for the item
        :return: QColor
        """

        return self.viewer().text_selected_color()

    def background_hover_color(self):
        """
        Returns the background color when the mouse is over the item
        :return: QColor
        """

        return self.viewer().background_hover_color()

    def background_selected_color(self):
        """
        Returns the background color when the item is selected
        :return: QColor
        """

        return self.viewer().background_selected_color()

    def rect(self):
        """
        Returns the rect for the current paint frame
        :return: QRect
        """

        return self._rect

    def set_rect(self, rect):
        """
        Sets the rect for the current paint frame
        :param rect: QRect
        """

        self._rect = rect

    def visual_rect(self, option):
        """
        Returns the visual rect for the item
        :param option: QStyleOptionViewItem
        :return: QRect
        """

        return QRect(option.rect)

    def icon_rect(self, option):
        """
        Returns the icon rect for the item
        :param option: QStyleOptionViewItem
        :return: QRect
        """

        padding = self.padding()
        rect = self.visual_rect(option)
        width = rect.width()
        height = rect.height()
        if self.is_label_under_item():
            height -= self.text_height()
        width -= padding
        height -= padding
        rect.setWidth(width)
        rect.setHeight(height)

        x = 0
        x += float(padding) / 2
        x += float((width - rect.width())) / 2

        y = float((height - rect.height())) / 2
        y += float(padding) / 2
        rect.translate(x, y)

        return rect

    def type_icon_rect(self, option):
        """
        Returns the type icon rect
        :param option:
        :return: QRect
        """

        padding = 2 * self.dpi()
        rect = self.icon_rect(option)
        x = rect.x() + padding
        y = rect.y() + padding
        rect = QRect(x, y, 13 * self.dpi(), 13 * self.dpi())

        return rect

    def paint_row(self, painter, option, index):
        """
        Paint performs low-level painting for the item
        :param painter: QPainter
        :param option: QStyleOptionViewItem
        :param index: QModelIndex
        """

        QTreeWidget.drawRow(self.treeWidget(), painter, option, index)

    def paint(self, painter, option, index):
        """
        Paint performs low-level painting for the item
        :param painter: QPainter
        :param option: QStyleOptionViewItem
        :param index: QModelIndex
        """

        self.set_rect(QRect(option.rect))
        painter.save()
        try:
            self.paint_background(painter, option, index)
            self.paint_icon(painter, option, index)
            if index.column() == 0 and self.blend_value() != 0:
                self.paint_blend_slider(painter, option, index)
            if self.is_text_visible():
                self.paint_text(painter, option, index)
            if index.column() == 0:
                self.paint_type_icon(painter, option)
            if index.column() == 0 and self.image_sequence():
                self.paint_playhead(painter, option)
        finally:
            painter.restore()

    def paint_background(self, painter, option, index):
        """
        Draw the background for the item
        :param painter: QPainter
        :param option: QStyleOptionViewItem
        :param index:QModelIndex
        """

        is_selected = option.state & QStyle.State_Selected
        is_mouse_over = option.state & QStyle.State_MouseOver
        painter.setPen(QPen(Qt.NoPen))
        visual_rect = self.visual_rect(option)
        if is_selected:
            color = self.background_selected_color()
        elif is_mouse_over:
            color = self.background_hover_color()
        else:
            color = self.backgroundColor()
        painter.setBrush(QBrush(color))

        if not self.viewer().is_icon_view():
            spacing = 1 * self.dpi()
            height = visual_rect.height() - spacing
            visual_rect.setHeight(height)

        painter.drawRect(visual_rect)

    def paint_icon(self, painter, option, index, align=None):
        """
        Draws the icon for the item
        :param painter: QPainter
        :param option: QStyleOptionViewItem
        :param index: int
        :param align: Qt.Align
        """

        column = index.column()
        pixmap = self.pixmap(column)
        if not pixmap:
            return

        rect = self.icon_rect(option)
        pixmap = self._scale_pixmap(pixmap, rect)
        pixmap_rect = QRect(rect)
        pixmap_rect.setWidth(pixmap.width())
        pixmap_rect.setHeight(pixmap.height())

        align = align or Qt.AlignHCenter | Qt.AlignVCenter

        x, y = 0, 0
        align_bottom_a = Qt.AlignBottom | Qt.AlignLeft
        align_bottom_b = align == Qt.AlignBottom | Qt.AlignHCenter or align == Qt.AlignBottom | Qt.AlignRight
        align_h_center_a = Qt.AlignHCenter or align == Qt.AlignCenter
        align_h_center_b = align == Qt.AlignHCenter | Qt.AlignBottom or align == Qt.AlignHCenter | Qt.AlignTop
        align_v_center_a = align == Qt.AlignVCenter | Qt.AlignLeft or align == Qt.AlignVCenter | Qt.AlignRight
        align_v_center_b = Qt.AlignVCenter or align == Qt.AlignCenter

        is_align_bottom = align == align_bottom_a or align_bottom_b
        is_align_h_center = align == align_h_center_a or align_h_center_b
        is_align_v_center = align == align_v_center_a or align_v_center_b
        if is_align_h_center:
            x += float(rect.width() - pixmap.width()) / 2
        elif is_align_v_center:
            y += float(rect.height() - pixmap.height()) / 2
        elif is_align_bottom:
            y += float(rect.height() - pixmap.height())

        pixmap_rect.translate(x, y)
        painter.drawPixmap(pixmap_rect, pixmap)

    def paint_text(self, painter, option, index):
        """
        Draws the text for the item
        :param painter: QPainter
        :param option: QStyleOptionViewItem
        :param index: int
        """

        column = index.column()
        if column == 0 and self.viewer().is_table_view():
            return

        self._paint_text(painter, option, column)

    def paint_type_icon(self, painter, option):
        """
        Draw the item type icon at the top left
        :param painter: QPainter
        :param option: QStyleOptionViewItem
        """

        rect = self.type_icon_rect(option)
        type_pixmap = self.type_pixmap()
        if type_pixmap:
            painter.setOpacity(0.5)
            painter.drawPixmap(rect, type_pixmap)
            painter.setOpacity(1)

    def paint_playhead(self, painter, option):
        """
        Pain the playhead if the item has an image sequence
        :param painter: QPainter
        :param option: QStyleOptionViewItem
        """

        image_sequence = self.image_sequence()
        if image_sequence and self.under_mouse():
            count = image_sequence.frame_count()
            current = image_sequence.current_frame_number()
            if count > 0:
                percent = float((count + current) + 1) / count - 1
            else:
                percent = 0

            icon_rect = self.icon_rect(option)
            playhead_color = self.playhead_color()

            painter.setPen(Qt.NoPen)
            painter.setBrush(QBrush(playhead_color))

            if percent <= 0:
                width = 0
            elif percent >= 1:
                width = icon_rect.width()
            else:
                width = (percent * icon_rect.width()) - 1

            height = 3 * self.dpi()
            y = icon_rect.y() + icon_rect.height() - (height - 1)

            painter.drawRect(icon_rect.x(), y, width, height)

    def paint_blend_slider(self, painter, option, index):
        if not self.PAINT_SLIDER or not self.viewer().is_icon_view():
            return

        painter.setPen(QPen(Qt.NoPen))
        rect = self.visual_rect(option)
        color = self.viewer().background_color().toRgb()
        color.setAlpha(75)
        painter.setBrush(QBrush(color))
        height = rect.height()
        ratio = self.blend_value()
        if ratio < 0:
            width = 0
        elif ratio > 100:
            width = rect.width()
        else:
            width = rect.width() * (float(ratio) / 100)
        rect.setWidth(width)
        rect.setHeight(height)

        rect = self.visual_rect(option)
        rect.setY(rect.y() + (4 * self.dpi()))
        color = self.viewer().text_color().toRgb()
        color.setAlpha(220)
        pen = QPen(color)
        align = Qt.AlignTop | Qt.AlignHCenter
        painter.setPen(pen)
        painter.drawText(rect, align, str(self.blend_value()) + "%")

    # =================================================================================================================
    # INTERNAL
    # =================================================================================================================

    def _thumbnail_from_image(self, image):
        """
        Called after the given image object has finished loading
        :param image: QImage
        """

        self.clear_cache()
        pixmap = QPixmap()
        pixmap.convertFromImage(image)
        icon = QIcon(pixmap)
        self._thumbnail_icon = icon
        if self.viewer():
            self.viewer().update()

    def _scale_pixmap(self, pixmap, rect):
        """
        Internal function that scales the given pixmap to given rect size
        The scaled pixmap is cached and its reused if its called with the same size
        :param pixmap: QPixmap
        :param rect: QRect
        :return: QPixmap
        """

        rect_changed = True

        if self._pixmap_rect:
            width_changed = self._pixmap_rect.width() != rect.width()
            height_changed = self._pixmap_rect.height() != rect.height()
            rect_changed = width_changed or height_changed

        if not self._pixmap_scaled or rect_changed:
            self._pixmap_scaled = pixmap.scaled(
                rect.width(), rect.height(), Qt.KeepAspectRatio, Qt.SmoothTransformation)
            self._pixmap_rect = rect

        return self._pixmap_scaled

    def _paint_text(self, painter, option, column):
        """
        Internal function used to paint the text
        :param painter: QPainter
        :param option: QStyleOption
        :param column: int
        """

        if self.viewer().is_icon_view():
            text = self.item.full_name() if self.item else self.name()
        else:
            label = self.label_from_column(column)
            text = self.display_text(label)

        color = self.text_color()
        is_selected = option.state & QStyle.State_Selected
        if is_selected:
            color = self.text_selected_color()

        visual_rect = self.visual_rect(option)
        width = visual_rect.width()
        height = visual_rect.height()
        padding = self.padding()
        x = padding / 2
        y = padding / 2
        visual_rect.translate(x, y)
        visual_rect.setWidth(width - padding)
        visual_rect.setHeight(height - padding)

        font = self.font(column)
        align = self.textAlignment(column)
        metrics = QFontMetrics(font)
        text_width = 1
        if text:
            text_width = metrics.width(text)
        if text_width > visual_rect.width() - padding:
            visual_width = visual_rect.width()
            text = metrics.elidedText(text, Qt.ElideMiddle, visual_width)
            align = Qt.AlignLeft

        align = align | Qt.AlignVCenter
        rect = QRect(visual_rect)

        if self.viewer().is_icon_view():
            if self.is_label_over_item() or self.is_label_under_item():
                padding = 8 if padding < 8 else padding
                height = metrics.height() + (padding / 2)
                y = (rect.y() + rect.height()) - height
                rect.setY(y)
                rect.setHeight(height)
            if self.is_label_over_item():
                color2 = self.viewer().background_color().toRgb()
                color2.setAlpha(200)
                painter.setPen(Qt.NoPen)
                painter.setBrush(QBrush(color2))
                painter.drawRect(rect)

        pen = QPen(color)
        painter.setPen(pen)
        painter.setFont(font)
        painter.drawText(rect, align, text)

    # ============================================================================================================
    # BLENDING
    # ============================================================================================================

    def is_blending_enabled(self):
        """
        Returns whether blending is enabled or not
        :return: bool
        """

        return self._blending_enabled

    def set_blending_enabled(self, flag):
        """
        Sets whether blending is enabled or not
        :param flag: bool
        """

        self._blending_enabled = flag

    def is_blending(self):
        """
        Returns whether blending is playing or not
        :return: bool
        """

        return self._blend_down

    def set_is_blending(self, flag):
        """
        Called when the middle mouse button is released
        :return: QMouseEvent
        """

        self._blend_down = flag
        if not flag:
            self._blend_position = None
            self._blend_prev_value = self.blend_value()

    def blend_value(self):
        """
        Returns the current blend value
        :return: float
        """

        return self._blend_value

    def set_blend_value(self, blend):
        """
        Sets the current blend value
        :param blend: float
        """

        if self.is_blending_enabled():
            self._blend_value = blend
            if self.PAINT_SLIDER:
                self.update()
            self.blendChanged.emit(blend)
            if self.PAINT_SLIDER:
                self.update()

    def blend_previous_value(self):
        """
        Returns the blend previous current value
        :return: float
        """

        return self._blend_prev_value

    def blend_position(self):
        """
        Returns current blend position
        :return: QPoint
        """

        return self._blend_position

    def blending_event(self, event):
        """
        Function that is called when the mouse moves while the middle mouse button is held down
        :param event: QMouseEvent
        """

        if self.is_blending():
            value = math.ceil((event.pos().x() - self.blend_position().x()) * 1.5) + self.blend_previous_value()
            try:
                self.set_blend_value(value)
            except Exception:
                self.set_is_blending(False)

    def reset_blending(self):
        """
        Resets the blending value to zero
        """

        self._blend_value = 0.0
        self._blend_prev_value = 0.09

    # =================================================================================================================
    # CALLBACKS
    # =================================================================================================================

    def _on_thumbnail_from_image(self, image):
        """
        Internal callback function that is called when an image object has finished loading
        """

        self.clear_cache()
        pixmap = QPixmap()
        pixmap.convertFromImage(image)
        icon = QIcon(pixmap)
        self._thumbnail_icon = icon
        if self.viewer():
            self.viewer().update()

    def _on_frame_changed(self):
        """
        Internal callback function that is triggered when the movei object updates to the given
        frame
        :return:
        """

        if not qtutils.is_control_modifier():
            self.update_frame()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains load widget for data items implementation for Maya
"""

from __future__ import print_function, division, absolute_import

import logging

from tpDcc.tools.datalibrary.widgets import load

LOGGER = logging.getLogger('tpDcc-tools-datalibrary')


class MayaLoadWidget(load.BaseLoadWidget):
    def __init__(self, client, item_view, *args, **kwargs):
        super(MayaLoadWidget, self).__init__(client=client, item_view=item_view, *args, **kwargs)

        item = self.item()
        if item:
            self._export_btn.setVisible(bool(item.functionality().get('export_data', False)))
            self._import_btn.setVisible(bool(item.functionality().get('import_data', False)))
            self._reference_btn.setVisible(bool(item.functionality().get('reference_data', False)))

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def ui(self):
        super(MayaLoadWidget, self).ui()

        # self._selection_set_button = buttons.BaseButton(parent=self)
        # self._selection_set_button.setIcon(resources.icon('group_objects'))

    def setup_signals(self):
        super(MayaLoadWidget, self).setup_signals()
        # self._selection_set_button.clicked.connect(self._on_show_selection_sets_menu)

    # ============================================================================================================
    # CALLBACKS
    # ============================================================================================================

    def _on_show_selection_sets_menu(self):
        """
        Internal callback function that shows selection sets menu
        """

        item = self.item()
        item.show_selection_sets_menu()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains base library item view implementation
"""

from __future__ import print_function, division, absolute_import

import os
import logging

from Qt.QtCore import Qt
from Qt.QtWidgets import QApplication, QAction

from tpDcc import dcc
from tpDcc.managers import resources
from tpDcc.libs.python import decorators, path as path_utils

from tpDcc.tools.datalibrary.core.views import dataitem

LOGGER = logging.getLogger('tpDcc-libs-datalibrary')


class _MetaBaseDataItem(type):

    def __call__(self, *args, **kwargs):
        as_class = kwargs.get('as_class', True)
        if dcc.client().is_maya():
            from tpDcc.libs.datalibrary.dccs.maya.core import dataitem
            if as_class:
                return dataitem.MayaDataItemView
            else:
                return type.__call__(dataitem.MayaDataItemView, *args, **kwargs)
        else:
            if as_class:
                return BaseDataItemView
            else:
                return type.__call__(BaseDataItemView, *args, **kwargs)


class BaseDataItemView(dataitem.DataItemView):

    NAME = 'Data View'

    def __init__(self, data_item=None, *args, **kwargs):
        super(BaseDataItemView, self).__init__(data_item=data_item, *args, **kwargs)

        # self._item.emitError.connect(self._on_item_error)

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def context_menu(self, menu, items=None):
        """
        Returns the context menu for the item
        :return: QMenu
        """

        self._select_content_action = QAction(resources.icon('cursor'), 'Select content', menu)
        self._select_content_action.triggered.connect(self.select_content)
        menu.addAction(self._select_content_action)
        menu.addSeparator()

    @classmethod
    def show_save_widget(cls, item_class, library_window, item_view=None):
        """
        Function used to show the create widget of the current item
        Override to allow us set the destination location for the save widget
        :param item_class: DataItem
        :param library_window: LibraryWindow
        :param item_view: LibraryItem or None
        """

        library = library_window.library()
        if not library:
            LOGGER.warning('Impossible to save data because library is not defined!')
            return

        save_widget_class = library_window.factory.get_save_widget_class(item_class)
        if not save_widget_class:
            LOGGER.warning(
                'Impossible to create new item of type "{}" because no save widget implementation defined!'.format(
                    item_class.__name__))
            return

        item_path = library_window.selected_folder_path()
        if not item_path or not os.path.exists(item_path):
            item_path = library_window.path()

        if not item_view:
            # NOTE: Here we generate a temporal file name to generate the proper composite data
            temp_item_path = path_utils.join_path(item_path, '__temp__{}'.format(item_class.EXTENSION))
            item_data = library.get(temp_item_path, only_extension=True)
            item_view_class = library_window.factory.get_view(item_data)
            item_view = item_view_class(item_data, library_window=library_window)

        widget = save_widget_class(item_view=item_view, client=library_window.client, parent=library_window)
        widget.set_folder_path(item_path)
        widget.set_library_window(library_window)
        library_window.set_save_widget(widget)
        library_window.folderSelectionChanged.connect(widget.set_folder_path)

    @classmethod
    def show_export_widget(cls, item_class, data_path, library_window):
        """
        Function used to show the export widget of the current item
        :param data_path: DataItem
        :param library_window: LibraryWindow
        """

        library = library_window.library()
        if not library:
            LOGGER.warning('Impossible to export data because library is not defined!')
            return

        export_widget_class = library_window.factory.get_export_widget_class(item_class)
        if not export_widget_class:
            LOGGER.warning(
                'Impossible to export item of type "{}" because no export widget implementation defined!'.format(
                    item_class.__name__))
            return

        if not data_path or not os.path.isfile(data_path):
            LOGGER.warning('Impossible to export item "{}" because its path does not exists: "{}"'.format(
                item_class, data_path))
            return

        item_data = library.get(data_path, only_extension=True)
        if not item_data:
            LOGGER.warning('Impossible to export data "{}" because was not found!'.format(data_path))
            return

        item_view_class = library_window.factory.get_view(item_data)
        item_view = item_view_class(item_data, library_window=library_window)

        widget = export_widget_class(item_view=item_view, client=library_window.client, parent=library_window)
        widget.set_library_window(library_window)
        library_window.set_export_widget(widget)

    def double_clicked(self):
        """
        Triggered when an item is double clicked
        """

        # self._item.load_from_current_values()

        load_function = self._item.functionality().get('load')
        if not load_function:
            return

        load_function()

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def select_content(self, **kwargs):
        """
        Select the contents of this item in the current DCC scene
        :param kwargs: dict
        """

        kwargs = kwargs or self._selection_modifiers()
        self._item.select_content(**kwargs)

    # ============================================================================================================
    # INTERNAL
    # ============================================================================================================

    def _selection_modifiers(self):
        """
        Internal function that returns a dictionary with the current key modifiers
        :return: dict
        """

        result = {'add': False, 'deselect': False}
        modifiers = QApplication.keyboardModifiers()

        if modifiers == Qt.ShiftModifier:
            result['deselect'] = True
        elif modifiers == Qt.ControlModifier:
            result['add'] = True

        return result

    def _on_item_error(self, error, trace):
        self.show_error_dialog(str(error), str(trace))


@decorators.add_metaclass(_MetaBaseDataItem)
class DataItemView(object):
    pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains library folder item widget implementation
"""

from __future__ import print_function, division, absolute_import

from Qt.QtCore import Qt, QLocale, QFileInfo
from Qt.QtWidgets import QDialogButtonBox, QAction
from Qt.QtGui import QPainter

import os

from tpDcc.managers import resources
from tpDcc.libs.python import path as path_utils
from tpDcc.libs.resources.core import theme, color as qt_color, pixmap as qt_pixmap, icon as qt_icon
from tpDcc.libs.qt.widgets import messagebox, action
from tpDcc.libs.datalibrary.data import folder

from tpDcc.tools.datalibrary.core import consts
from tpDcc.tools.datalibrary.widgets import preview
from tpDcc.tools.datalibrary.core.views import dataitem


@theme.mixin
class FolderItemView(dataitem.DataItemView):

    NAME = 'Folder View'
    REPRESENTING = [folder.FolderData.__name__]

    LOAD_WIDGET_CLASS = preview.PreviewWidget

    DEFAULT_THUMBNAIL_NAME = 'folder.png'

    _THUMBNAIL_ICON_CACHE = dict()

    def _init__(self, *args, **kwargs):
        super(FolderItemView, self).__init__(*args, **kwargs)

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    @classmethod
    def show_save_widget(cls, item_class, library_window, item_view=None):
        """
        Function used to show the create widget of the current item
        :param library_window: LibraryWindow
        :param item_view: LibraryItem or None
        """

        path = library_window.selected_folder_path() or library_window.path()
        name, button = messagebox.MessageBox.input(
            library_window, 'Create Folder', 'Create a new folder with the name:')
        if not name or not button == QDialogButtonBox.Ok:
            return
        path = path_utils.join_path(path, name.strip())

        item_data = item_class(path, library_window.library())
        if not item_data:
            return

        save_function = item_data.functionality().get('save')
        if not save_function:
            return

        valid = save_function()
        if not valid:
            return

        if library_window:
            library_window.sync()
            library_window.refresh()
            library_window.select_folder_path(path)

    def double_clicked(self):
        """
        Triggered when an item is double clicked
        """

        if not self.library_window():
            return

        self.library_window().select_folder_path(self.path())

    def thumbnail_icon(self):
        """
        Returns the thumbnail icon
        :return: QIcon
        """

        # custom_path = self.custom_icon_path()
        # if custom_path and '/' not in custom_path and '\\' not in custom_path:
        #     custom_path = resources.icon('icons/{}'.format(self.theme().style()), custom_path)
        # if not custom_path or not os.path.isfile(custom_path):
        #     return super(FolderItemView, self).thumbnail_icon()

        return super(FolderItemView, self).thumbnail_icon()

        color = self.icon_color()
        if not color:
            color = consts.DEFAULT_FOLDER_ICON_COLOR

        icon_key = custom_path + color

        icon = self._THUMBNAIL_ICON_CACHE.get(icon_key)
        if not icon:
            color1 = qt_color.Color.from_string(color)
            color2 = qt_color.Color.from_string('rgb(255, 255, 255, 150)')
            pixmap1 = qt_pixmap.Pixmap(self.THUMBNAIL_PATH)
            pixmap2 = qt_pixmap.Pixmap(custom_path)
            pixmap1.set_color(color1)
            pixmap2.set_color(color2)
            pixmap1 = pixmap1.scaled(128, 128, Qt.KeepAspectRatio, Qt.SmoothTransformation)
            pixmap2 = pixmap2.scaled(64, 64, Qt.KeepAspectRatio, Qt.SmoothTransformation)
            x = (128 - pixmap2.width()) / 2
            y = (128 - pixmap2.width()) / 2
            painter = QPainter(pixmap1)
            painter.drawPixmap(x, y + 5, pixmap2)
            painter.end()
            icon = qt_icon.Icon(pixmap1)
            self._THUMBNAIL_ICON_CACHE[icon_key] = icon

        return self._THUMBNAIL_ICON_CACHE.get(icon_key)

    def context_edit_menu(self, menu, items=None):
        """
        This function is called when the user opens context menu
        The given menu is shown as a submenu of the main context menu
        This function can be override to create custom context menus in LibraryItems
        :param menu: QMenu
        :param items: list(LibraryItem)
        """

        super(FolderItemView, self).context_edit_menu(menu, items=items)

        # show_preview_action = QAction(resources.icon('preview'), 'Show in Preview', menu)
        # show_preview_action.triggered.connect(self._on_show_preview_from_menu)
        # menu.addAction(show_preview_action)
        # menu.addSeparator()
        # color_picker_action = action.ColorPickerAction(menu)
        # color_picker_action.picker().set_colors(consts.DEFAULT_FOLDER_ICON_COLORS)
        # color_picker_action.picker().colorChanged.connect(self.set_icon_color)
        # color_picker_action.picker().set_current_color(self.icon_color())
        # menu.addAction(color_picker_action)
        # icon_name = self.item.data.get('icon', '')
        # icon_picker_action = action.IconPickerAction(menu)
        # icon_picker_action.picker().set_icons(consts.DEFAULT_FOLDER_ICONS)
        # icon_picker_action.picker().set_current_icon(icon_name)
        # icon_picker_action.picker().iconChanged.connect(self.set_custom_icon)
        # menu.addAction(icon_picker_action)

        # color_picker_action.setVisible(False)
        # icon_picker_action.setVisible(False)

    def create_overwrite_menu(self, menu):
        """
        Overwrites base LibraryDataItem create_overwrite_menu function to hide the overwrite menu action
        :param menu: QMenu
        """

        pass

    def _on_metadata_updated(self, metadata):
        """
        Internal callback function that is called when metadata is updated
        Override to support the update of the library widget directly
        :param metadata: dict
        """

        super(FolderItemView, self)._on_metadata_updated(metadata)

        if self.library_window():
            self.library_window().set_folder_data(self.path(), self.item.data)
            self.update_icon()

    # ============================================================================================================
    # CALLBACK
    # ============================================================================================================

    def _on_show_preview_from_menu(self):
        if self.library_window():
            self.library_window().viewer().clear_selection()
        self.show_preview_widget(self.library_window())



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains search data library widget implementation
"""

from __future__ import print_function, division, absolute_import

import logging
from functools import partial

from Qt.QtCore import Qt, Signal, QSize
from Qt.QtWidgets import QStyle, QLineEdit, QMenu, QAction
from Qt.QtGui import QCursor

from tpDcc.managers import resources
from tpDcc.libs.qt.widgets import buttons

LOGGER = logging.getLogger('tpDcc-libs-datalibrary')


class DataSearcherWidget(QLineEdit):

    SPACE_OPEARTOR = 'and'
    PLACEHOLDER_TEXT = 'Search'

    searchChanged = Signal()

    def __init__(self, parent=None):
        super(DataSearcherWidget, self).__init__(parent=parent)

        self._library = None
        self._space_operator = 'and'

        self._icon_btn = buttons.BaseButton(parent=self)
        self._icon_btn.setIcon(resources.icon('search'))
        self._icon_btn.setObjectName('searchButton')
        self._icon_btn.setIconSize(QSize(12, 12))
        self._icon_btn.clicked.connect(self._on_icon_clicked)
        self._clear_btn = buttons.BaseButton(parent=self)
        self._clear_btn.setIcon(resources.icon('delete'))
        self._clear_btn.setObjectName('clearButton')
        self._clear_btn.setIconSize(QSize(12, 12))
        self._clear_btn.setCursor(Qt.ArrowCursor)
        self._clear_btn.setToolTip('Clear all search text')
        self._clear_btn.clicked.connect(self._on_clear_clicked)
        self.setPlaceholderText(self.PLACEHOLDER_TEXT)
        self.textChanged.connect(self._on_text_changed)
        self.update()

        tip = 'Search all current items.'
        self.setToolTip(tip)
        self.setStatusTip(tip)

        self._icon_btn.setStyleSheet('background-color: transparent')
        self._clear_btn.setStyleSheet('background-color: transparent')
        # self.setStyleSheet('border-radius: 13px; border: 2px;')

    # ============================================================================================================
    # OVERRIDES
    # ============================================================================================================

    def update(self):
        """
        Overrides base QLineEdit update function
        """

        self.update_clear_button()

    def resizeEvent(self, event):
        """
        Overrides base QLineEdit resizeEvent function
        :param event: QResizeEvent
        """

        super(DataSearcherWidget, self).resizeEvent(event)

        self.setTextMargins(self.height(), 0, 0, 0)
        size = QSize(self.height(), self.height())
        self._icon_btn.setIconSize(size)
        self._icon_btn.setFixedSize(size)
        self._clear_btn.setIconSize(size)
        x = self.width() - self.height()
        self._clear_btn.setGeometry(x, 0, self.height(), self.height())

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Escape:
            self.clear()
        super(DataSearcherWidget, self).keyPressEvent(event)

    def contextMenuEvent(self, event):
        self.show_context_menu()

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def library(self):
        """
        Returns the library model for the menu
        :return: Library
        """

        return self._library

    def set_library(self, library):
        """
        Set the library model for the menu
        :param library: Library
        """

        self._library = library

    def space_operator(self):
        """
        Returns the space operator for the search widget
        :return: str
        """

        return self._space_operator

    def set_space_operator(self, space_operator):
        """
        Sets the space operator
        :param space_operator: str
        """

        self._space_operator = space_operator
        self.search()

    def search(self):
        """
        Run the search query on the library
        """

        if self.library():
            self.library().add_query(self.query())
            self.library().search()
        else:
            LOGGER.info('No library found for the search widget')

        self.update_clear_button()
        self.searchChanged.emit()

    def query(self):
        """
        Returns the query used for the library
        :return: dict
        """

        text = str(self.text())
        filters = list()
        for filter_ in text.split(' '):
            if filter_.split():
                filters.append(('identifier', 'contains', filter_))
        unique_name = 'searchWidget' + str(id(self))

        return {'name': unique_name, 'operator': self.space_operator(), 'filters': filters}

    def update_clear_button(self):
        """
        Updates the clear button depending on the current text
        """

        text = self.text()
        if text:
            self._clear_btn.show()
        else:
            self._clear_btn.hide()

    # ============================================================================================================
    # MENU
    # ============================================================================================================

    def show_context_menu(self):
        """
        Creates and shows the context menu for the search widget
        :return: QAction
        """

        menu = QMenu(self)
        standard_menu = self.createStandardContextMenu()
        standard_menu.setTitle('Edit')
        menu.addMenu(standard_menu)

        sub_menu = QMenu(menu)
        sub_menu.setTitle('Space Operator')
        menu.addMenu(sub_menu)

        or_action = QAction('OR', menu)
        or_action.setCheckable(True)
        or_callback = partial(self.set_space_operator, 'or')
        or_action.triggered.connect(or_callback)
        if self.space_operator() == 'or':
            or_action.setChecked(True)
        sub_menu.addAction(or_action)

        and_action = QAction('AND', menu)
        and_action.setCheckable(True)
        and_callback = partial(self.set_space_operator, 'and')
        and_action.triggered.connect(and_callback)
        if self.space_operator() == 'and':
            and_action.setChecked(True)
        sub_menu.addAction(and_action)

        action = menu.exec_(QCursor.pos())

        return action

        # ============================================================================================================
    # SETTINGS
    # ============================================================================================================

    def settings(self):
        """
        Returns a dictionary of the current widget state
        :return: dict
        """

        settings = {
            'text': self.text(),
            'spaceOperator': self.space_operator()
        }

        return settings

    def set_settings(self, settings):
        """
        Restore the widget state from a settings dictonary
        :param settings: dict
        """

        text = settings.get('text', '')
        self.setText(text)
        space_operator = settings.get('spaceOperator')
        if space_operator:
            self.set_space_operator(space_operator)

    # ============================================================================================================
    # INTERNAL
    # ============================================================================================================

    def _search_line_frame_width(self):
        return self.style().pixelMetric(QStyle.PM_DefaultFrameWidth)

    def _clear_button_padded_width(self):
        return self._clear_btn.width() + self._search_line_frame_width() * 2

    def _clear_button_padded_height(self):
        return self._clear_btn.height() + self._search_line_frame_width() * 2

    def _search_button_padded_width(self):
        return self._icon_btn.width() + self._search_line_frame_width() * 2

    def _search_button_padded_height(self):
        return self._icon_btn.height() + self._search_line_frame_width() * 2

    # ============================================================================================================
    # CALLBACKS
    # ============================================================================================================

    def _on_icon_clicked(self):
        """
        Internal callback function that is triggered when the user clcks on the search icon
        """

        if not self.hasFocus():
            self.setFocus()

    def _on_clear_clicked(self):
        """
        Internal callback function that is triggered when the user clicks the cross icon
        """

        self.setText('')
        self.setFocus()

    def _on_text_changed(self):
        """
        Internal callback function that is triggered when the text changes
        """

        self.search()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Module that contains Maya sets menu implementation
"""

from __future__ import print_function, division, absolute_import

import os
from functools import partial

from Qt.QtWidgets import QMenu, QAction
from Qt.QtGui import QCursor

from tpDcc.managers import resources
from tpDcc.tools.datalibrary.core import utils
# from tpDcc.libs.datalibrary.dccs.maya.data import setsitem


class SetsMenu(QMenu):
    def __init__(self, item, parent=None, namespaces=None, enable_select_content=True):
        super(SetsMenu, self).__init__('Selection Sets', parent or item.library_window())

        self._item = item
        self._namespaces = namespaces
        self._enable_select_content = enable_select_content

        # icon = resources.icon(setsitem.SetsItem.ICON_NAME)
        # self.setIcon(icon)

        self.reload()

    # ============================================================================================================
    # CLASS METHODS
    # ============================================================================================================

    @classmethod
    def from_path(cls, path, parent=None, library_window=None, **kwargs):
        """
        Returns a new SetsMenu from the given path
        :param path: str
        :param parent: QMenu or None
        :param library_window: LibraryWindow or None
        :param kwargs: dict
        :return: QAction
        """

        item = setsitem.SetsItem(path, library_window=library_window)
        return cls(item, parent, enable_select_content=False, **kwargs)

    @classmethod
    def show_sets_menu(cls, path, **kwargs):
        """
        Shows selection set menu at the current cursor position
        :param path: str
        :param kwargs: dict
        """

        menu = cls.from_path(path, **kwargs)
        pos = QCursor.pos()

        return menu.exec_(pos)

    # ============================================================================================================
    # BASE
    # ============================================================================================================

    def item(self):
        """
        Returns current item
        :return: MayaDataItem
        """

        return self._item

    def namespaces(self):
        """
        Returns item namespaces
        :return: list(str)
        """

        return self._namespaces

    def select_content(self):
        """
        Select item objects
        """

        self.item().select_content(namespaces=self.namespaces())

    def selection_sets(self):
        """
        :return: list(SetsItem)
        """

        def _match(path):
            return path.endswith('.set')

        items = list()
        library_window = self.item().library_window()

        # path = self.item().path()
        # paths = utils.walkup(path, match=_match, depth=10)
        # paths = list(paths)
        # for path in paths:
        #     item = setsitem.SetsItem(path)
        #     item.set_library_window(library_window)
        #     items.append(item)

        return items

    def reload(self):
        """
        Reloads current menu data
        """

        self.clear()

        if self._enable_select_content:
            select_content_action = QAction(resources.icon('cursor'), 'Select content', self)
            select_content_action.triggered.connect(self.item().select_content)
            self.addAction(select_content_action)
            self.addSeparator()

        selection_sets = self.selection_sets()
        if not selection_sets:
            action = QAction('No selection sets found!', self)
            action.setEnabled(False)
            self.addAction(action)
            return

        for selection_set in selection_sets:
            dirname = os.path.basename(os.path.dirname(selection_set.path()))
            basename = os.path.basename(selection_set.path()).replace(selection_set.EXTENSION, '')
            nice_name = '{}: {}'.format(dirname, basename)
            selection_set_action = QAction(nice_name, self)
            selection_set_action_callback = partial(selection_set.load, namespaces=self.namespaces())
            selection_set_action.triggered.connect(selection_set_action_callback)
            self.addAction(selection_set_action)
